# Copyright (C) 2002, 2003 Intevation GmbH <intevation@intevation.de>
# Author: Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL coming with the software for details.

"""Convert Well-Known Text format to python objects

Main entry point is the convert_well_known_text function which takes a
geometry in Well-Known Text format and returns a python object with the
geometry.
"""
import re

__version__ = "$Revision: 1.1 $"
# $Source: /thubanrepository/thuban/Thuban/Model/wellknowntext.py,v $
# $Id: wellknowntext.py,v 1.1 2003/08/19 11:00:40 bh Exp $

_open_parens = r"[ \t]*(\([ \t]*)*"
_close_parens = r"[ \t]*(\)[ \t]*)+"
rx_point_list = re.compile(
    _open_parens + r"(?P<coords>[^\)]+)" + _close_parens + ",?"
)


def parse_coordinate_lists(wkt):
    """Return the coordinates in wkt as a list of lists of coordinate pairs.

    The wkt parameter is the coordinate part of a geometry in well-known
    text format.
    """
    geometry = []
    while wkt:
        match = rx_point_list.match(wkt)
        if match:
            poly = []
            wktcoords = match.group("coords")
            for pair in wktcoords.split(","):
                # a pair may be a triple actually. For now we just
                # ignore any third value
                x, y = list(map(float, pair.split()))[:2]
                poly.append((x, y))
            geometry.append(poly)
            wkt = wkt[match.end(0) :].strip()
        else:
            raise ValueError("Invalid well-known-text (WKT) syntax")
    return geometry


def parse_multipolygon(wkt):
    """
    Return the MULTIPOLYGON geometry wkt as a list of lists of float pairs
    """
    return parse_coordinate_lists(wkt)


def parse_polygon(wkt):
    """Return the POLYGON geometry in wkt as a list of float pairs"""
    return parse_coordinate_lists(wkt)


def parse_multilinestring(wkt):
    """
    Return the MULTILINESTRING geometry wkt as a list of lists of float pairs
    """
    return parse_coordinate_lists(wkt)


def parse_linestring(wkt):
    """Return the LINESTRING geometry in wkt as a list of float pairs"""
    return parse_coordinate_lists(wkt)[0]


def parse_point(wkt):
    """Return the POINT geometry in wkt format as pair of floats"""
    return parse_coordinate_lists(wkt)[0][0]


# map geometry types to parser functions
_function_map = [
    ("MULTIPOLYGON", parse_multipolygon),
    ("POLYGON", parse_polygon),
    ("MULTILINESTRING", parse_multilinestring),
    ("LINESTRING", parse_linestring),
    ("POINT", parse_point),
]


def convert_well_known_text(wkt):
    """Return the geometry given in well-known text format as python objects

    The function accepts only 2D data and supports the POINT, POLYGON,
    MULTIPOLYGON, LINESTRING and MULTILINESTRING geometries.

    The structure of the return value depends on the geometry type. For
    MULTIPOLYGON and MULTILINESTRING return a list of lists of
    coordinate pairs. For POLYGON and LINESTRING return a list of
    coordinate pairs. For POINT return a coordinate pair. All
    coordinates are floats.

    The string wkt may contain an SRID specification in addition to the
    actual geometry. This SRID is ignored.
    """
    parts = wkt.split(";")
    if not parts:
        raise ValueError("No recognized geometry in WKT string")
    for part in parts:
        part = part.strip()
        if part.startswith("SRID"):
            # ignore SRIDs
            continue
        for geotype, function in _function_map:
            if part.startswith(geotype):
                return function(part[len(geotype) :])
        raise ValueError("Unsupported WKT-part %s" % repr(part[:20]))



# -*- coding: utf-8 -*-
# pylint: disable=import-outside-toplevel,too-many-lines
"""Plotting utility for generating maps, windroses and everything else under
the sun.

This module provides a wrapper around `Basemap` and `windrose` packages.  It
tries to be general so to work for others, but may contain some unfortunate
hard coded values.  Bad daryl!

Example:
    Here is a basic example of usage.

    >>> from pyiem.plot.geoplot import MapPlot
    >>> m = MapPlot(sector='conus', title='My Fancy Title')
    >>> m.postprocess(filename='myplot.png')
    >>> m.close()

"""
# stdlib
from io import BytesIO
import tempfile
import os
import sys
import subprocess
import shutil
import pickle
import datetime
import warnings

#
import requests
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import shape

#
from scipy.signal import convolve2d
from scipy.interpolate import NearestNDInterpolator

#
from PIL import Image

# Matplotlib
from matplotlib.patches import Polygon
import matplotlib.colors as mpcolors
from matplotlib.patches import Wedge
import matplotlib.colorbar as mpcolorbar
import matplotlib.patheffects as PathEffects

# cartopy
import cartopy
import cartopy.crs as ccrs
import cartopy.feature as cfeature

from pyiem.plot.use_agg import plt
from pyiem.plot.util import (
    sector_setter,
    mask_outside_polygon,
    polygon_fill,
    mask_outside_geom,
    draw_logo,
    fitbox,
)
from pyiem.reference import (  # noqa: F401  # pylint: disable=unused-import
    Z_CF,
    Z_FILL,
    Z_FILL_LABEL,
    Z_CLIP,
    Z_CLIP2,
    Z_POLITICAL,
    Z_OVERLAY,
    Z_OVERLAY2,
    Z_FRAME,
)
from pyiem.util import ssw, LOG
from pyiem.datatypes import speed, direction
from pyiem.plot.colormaps import stretch_cmap
import pyiem.meteorology as meteorology

# we ran the cartopy/tools downloader
cartopy.config["pre_existing_data_dir"] = "/opt/miniconda3/cartopy_data/"
# Set a saner default for apache et al
cartopy.config["data_dir"] = "/tmp/"


DATADIR = os.sep.join([os.path.dirname(__file__), "..", "data"])
MAIN_AX_BOUNDS = [0.01, 0.05, 0.898, 0.85]
CAX_BOUNDS = [0.917, 0.1, 0.02, 0.8]


def true_filter(_bm, _key, _val):
    """Always return true"""
    return True


def cwa_filter(bm, _key, val):
    """A filter for checking a key against current plot"""
    return val.get("cwa", "") == bm.cwa


def state_filter(bm, key, _val):
    """A filter for checking a key against current plot"""
    return key[:2] == bm.state


def load_bounds(filebase):
    """Load bounds file

    Args:
      filebase (str): the basename of the file containing the data

    Returns:
      numpy 2d array of the data
    """
    fn = "%s/%s.npy" % (DATADIR, filebase)
    if not os.path.isfile(fn):
        LOG.info("load_bounds(%s) is missing!", fn)
        return
    return np.load(fn)


def load_pickle_pd(filename):
    """Load a pickled pandas dataframe

    Args:
      filename(str): The filename to load, relative to project data/

    Returns:
      pandas.DataFrame
    """
    fn = "%s/%s" % (DATADIR, filename)
    if not os.path.isfile(fn):
        LOG.info("load_pickle_pd(%s) failed, file is missing", fn)
        return
    return pd.read_pickle(fn)


def load_pickle_geo(filename):
    """Load pickled dictionaries containing geometries and other metadata

    Args:
      filename(str): The filename to load, contained in project data/

    Returns:
      dict: The dictionary of data
    """
    fn = "%s/%s" % (DATADIR, filename)
    if not os.path.isfile(fn):
        LOG.info("load_pickle_geo(%s) failed, file is missing!", fn)
        return dict()
    pickle_opts = dict()
    if sys.version_info.major > 2:
        pickle_opts["encoding"] = "bytes"
    return pickle.load(open(fn, "rb"), **pickle_opts)


class MapPlot:
    """An object representing a cartopy plot.

    An object that allows one to quickly and easily generate map plots of data
    with some customization possible.  This is what drives most of the plots
    you see on the IEM website.

    Example:
      Here is an example of usage::

        mplot = MapPlot(sector='midwest', title='My Plot Title')
        mplot.plot_values([-99,-95], [44,45], ['hi','there'])
        mplot.postprocess(filename='test.png')
        mplot.close()

    Attributes:
        fig (matplotlib.Figure): figure object
        ax (matplotlib.Axes): main figure plot axes

    """

    def __init__(self, sector="iowa", figsize=(10.24, 7.68), **kwargs):
        """Construct a MapPlot

        Args:
          sector (str): plot domain, set 'custom' to bring your own projection
          kwargs:
            projection (cartopy.crs,optional): bring your own projection
            north (float,optional): Plot top bounds (degN Lat)
            south (float,optional): Plot bottom bounds (degN Lat)
            east (float,optional): Plot right bounds (degE Lon)
            west (float,optional): Plot left bounds (degE Lon)
            titlefontsize (int): fontsize to use for the plot title
            subtitlefontsize (int): fontsize to use for the plot subtitle
            continentalcolor (color): color to use for continental coloring
            debug (bool): enable debugging
            aspect (str): plot aspect, defaults to equal
            fig (matplotlib.pyplot.figure,optional): provide a figure instance
              for more advanced plot control.
            logo (str,optional): logo name to slap on the plot.
        """
        self.debug = kwargs.get("debug", False)
        self.fig = kwargs.get("fig")
        if self.fig is None:
            self.fig = plt.figure(
                num=None, figsize=figsize, dpi=kwargs.get("dpi", 100)
            )
        # Storage of axes within this plot
        self.state = None
        self.cwa = None
        self.textmask = None  # For our plot_values magic, to prevent overlap
        self.sector = sector
        self.cax = plt.axes(CAX_BOUNDS, frameon=False, yticks=[], xticks=[])
        self.axes = []
        self.ax = None
        self.pr_ax = None
        self.hi_ax = None
        self.ak_ax = None
        # hack around sector=iowa
        if self.sector == "iowa":
            self.sector = "state"
            self.state = "IA"
        sector_setter(self, MAIN_AX_BOUNDS, **kwargs)

        for _a in self.axes:
            if _a is None:
                continue
            # legacy usage of axisbg here
            _c = kwargs.get(
                "axisbg", kwargs.get("continentalcolor", "#EEEEEE")
            )
            _a.add_feature(cfeature.LAND, facecolor=_c, zorder=Z_CF)
            _a.add_feature(cfeature.COASTLINE, lw=1.0, zorder=Z_POLITICAL)
            _a.add_feature(cfeature.BORDERS, lw=1.0, zorder=Z_POLITICAL)
            _a.add_feature(
                cfeature.LAKES, facecolor=(0.4471, 0.6235, 0.8117), zorder=Z_CF
            )
            if "nostates" not in kwargs:
                states = load_pickle_geo("us_states.pickle")
                _a.add_geometries(
                    [val["geom"] for key, val in states.items()],
                    crs=ccrs.PlateCarree(),
                    lw=1.0,
                    edgecolor=kwargs.get("statebordercolor", "k"),
                    facecolor="None",
                    zorder=Z_POLITICAL,
                )

        if kwargs.get("logo") == "dep":
            draw_logo(self.fig, "deplogo.png")
        elif not kwargs.get("nologo"):
            draw_logo(self.fig, "logo.png")
        if "title" in kwargs:
            fitbox(
                self.fig,
                kwargs.get("title"),
                0.09 if not kwargs.get("nologo") else 0.02,
                0.99,
                0.94,
                0.99,
                textsize=kwargs.get("titlefontsize", 18),
            )
        if "subtitle" in kwargs:
            fitbox(
                self.fig,
                kwargs.get("subtitle"),
                0.09 if not kwargs.get("nologo") else 0.02,
                0.99,
                0.91,
                0.94,
                textsize=kwargs.get("subtitlefontsize", 12),
            )

        if "nocaption" not in kwargs:
            self.fig.text(
                0.01,
                0.03,
                ("%s :: generated %s")
                % (
                    kwargs.get("caption", "Iowa Environmental Mesonet"),
                    datetime.datetime.now().strftime("%d %B %Y %I:%M %p %Z"),
                ),
            )

    def close(self):
        """ Close the figure in the case of batch processing """
        plt.close()

    def draw_usdm(self, valid=None, filled=True, hatched=False):
        """Overlay the US Drought Monitor

        This utilizes a GeoJSON web service provided by the IEM.  The provided
        date to this method is passed to the web service which rectifies the
        date to provide the USDM analysis valid for that date.  If no date
        is specified, the current analysis is plotted.

        Args:
          valid (str or datetime.date): The valid time to plot this USDM
          filled (boolean): Should we draw lines or filled polygons
          hatched (boolean): Should we use hatch filling

        Return:
          date that the USDM is valid for
        """
        colors = ["#ffff00", "#fcd37f", "#ffaa00", "#e60000", "#730000"]
        hatches = ["+", "/", "\\", "-", "x"]
        if valid is None:
            valid = ""
        elif isinstance(valid, datetime.date):
            valid = valid.strftime("%Y-%m-%d")
        elif isinstance(valid, datetime.datetime):
            valid = valid.strftime("%Y-%m-%d")
        url = ("http://mesonet.agron.iastate.edu/geojson/usdm.py?date=%s") % (
            valid,
        )
        try:
            req = requests.get(url, timeout=30)
        except requests.ConnectionError as exp:
            warnings.warn("draw_usdm IEM USDM Webservice failed: %s" % (exp,))
            return None
        df = gpd.GeoDataFrame().from_features(req.json())
        lw = 1 if filled else 4.0
        usdm_valid = None
        for _, row in df.iterrows():
            color = colors[row["dm"]]
            geom = shape(row["geometry"])
            usdm_valid = row["date"]
            fc = color if filled else "None"
            ec = color if not filled else "k"
            self.ax.add_geometries(
                [geom],
                ccrs.PlateCarree(),
                facecolor="None",
                edgecolor=ec,
                linewidth=lw,
                zorder=Z_OVERLAY2,
            )
            if filled:
                self.ax.add_geometries(
                    [geom],
                    ccrs.PlateCarree(),
                    facecolor=fc,
                    alpha=0.5,
                    edgecolor="None",
                    zorder=Z_OVERLAY,
                )
            elif hatched:
                self.ax.add_geometries(
                    [geom],
                    ccrs.PlateCarree(),
                    facecolor="None",
                    hatch=hatches[row["dm"]],
                    edgecolor=color,
                    zorder=Z_OVERLAY2 + 2,
                )

        if usdm_valid is not None:
            self.ax.text(
                0.99,
                0.99,
                "D4",
                color="k",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color=colors[4]),
                zorder=Z_OVERLAY2 + 3,
            )
            self.ax.text(
                0.955,
                0.99,
                "D3",
                color="k",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color=colors[3]),
                zorder=Z_OVERLAY2 + 3,
            )
            self.ax.text(
                0.92,
                0.99,
                "D2",
                color="k",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color=colors[2]),
                zorder=Z_OVERLAY2 + 3,
            )
            self.ax.text(
                0.885,
                0.99,
                "D1",
                color="k",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color=colors[1]),
                zorder=Z_OVERLAY2 + 3,
            )
            self.ax.text(
                0.85,
                0.99,
                "D0",
                color="k",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color=colors[0]),
                zorder=Z_OVERLAY2 + 3,
            )
            self.ax.text(
                0.815,
                0.99,
                "USDM %s" % (usdm_valid,),
                color="w",
                transform=self.ax.transAxes,
                va="top",
                ha="right",
                bbox=dict(color="k"),
                zorder=Z_OVERLAY2 + 3,
            )
            return datetime.datetime.strptime(usdm_valid, "%Y-%m-%d")
        return None

    def draw_colorbar(self, clevs, cmap, norm, **kwargs):
        """Draw the colorbar on the structed plot using `self.cax`.

        Args:
          clevs (list): The levels used in the classification
          cmap (matplotlib.colormap): The colormap
          norm (normalize): The value normalizer
          title (str,optional): Place a label on the side, adjusts the plot
            accordingly to allow this text to fit, no multiline please!
          spacing (str,optional): should the colorbar be `uniform` or
            `proportional`, defaults to `uniform`
        """
        if self.cax is None:
            return
        extend = kwargs.get("extend", "both")
        # Do we need to muck with the cmap
        # pylint: disable=W0212
        if (extend in ["max", "both"] and cmap._rgba_over is None) or (
            extend in ["min", "both"] and cmap._rgba_under is None
        ):
            cmap = stretch_cmap(cmap, clevs, extend=extend)
            norm = mpcolors.BoundaryNorm(clevs, cmap.N)

        clevlabels = kwargs.get("clevlabels", clevs)

        under = clevs[0] - (clevs[1] - clevs[0])
        over = clevs[-1] + (clevs[-1] - clevs[-2])
        # create some faked levels in case we want an extended colorbar
        if extend == "min":
            blevels = np.concatenate([[under], clevs])
        elif extend == "max":
            blevels = np.concatenate([clevs, [over]])
        elif extend == "both":
            blevels = np.concatenate([[under], clevs, [over]])
        else:
            blevels = clevs
        stride = slice(None, None, int(kwargs.get("clevstride", 1)))
        cb2 = mpcolorbar.ColorbarBase(
            self.cax,
            cmap=cmap,
            norm=norm,
            boundaries=blevels,
            extend=extend,
            ticks=clevs[stride],
            spacing=kwargs.get("spacing", "uniform"),
            orientation="vertical",
        )

        def _myrepr(val):
            """avoid list conversion in matplotlib that fowls numpy floats."""
            try:
                return "%g" % (val,)
            except TypeError:
                return "%s" % (val,)

        cb2.ax.set_yticklabels(list(map(_myrepr, clevlabels[stride])))
        # Attempt to quell offset that sometimes appears with large numbers
        cb2.ax.get_yaxis().get_major_formatter().set_offset_string("")
        for label in cb2.ax.get_yticklabels():
            sz = len(label.get_text())
            if sz > 4:
                label.set_fontsize(12)
            else:
                label.set_fontsize(16)
            if sz > 6:
                label.set_rotation(45)

        if "units" in kwargs:
            self.fig.text(
                0.99, 0.03, "data units :: %s" % (kwargs["units"],), ha="right"
            )

        title = kwargs.get("title")
        if title:
            self.ax.set_position(
                [
                    MAIN_AX_BOUNDS[0],
                    MAIN_AX_BOUNDS[1],
                    MAIN_AX_BOUNDS[2] - 0.03,
                    MAIN_AX_BOUNDS[3],
                ]
            )
            cb2.ax.text(
                -0.05,
                0.5,
                title,
                rotation=90,
                fontsize=16,
                transform=cb2.ax.transAxes,
                ha="right",
                va="center",
            )

    def plot_station(self, data, **kwargs):
        """Plot values on a map in a station plot like manner.

        Args:
          data (list): list of dicts with station data to plot
          fontsize (int): font size to use for plotted text
        """
        (x0, x1) = self.ax.set_xlim()
        # size to use for circles
        circlesz = (x1 - x0) / 180.0
        # (y0, y1) = self.ax.set_ylim()
        offsets = {
            1: [-4, 4, "right", "bottom"],
            2: [0, 4, "center", "bottom"],
            3: [4, 4, "left", "bottom"],
            4: [-4, 0, "right", "center"],
            5: [0, 0, "center", "center"],
            6: [4, 0, "left", "center"],
            7: [-4, -4, "right", "top"],
            8: [0, -4, "center", "top"],
            9: [4, -4, "left", "top"],
        }

        mask = np.zeros(self.fig.canvas.get_width_height(), bool)
        for stdata in data:
            (x, y) = self.ax.projection.transform_point(
                stdata["lon"], stdata["lat"], ccrs.Geodetic()
            )
            (imgx, imgy) = self.ax.transData.transform([x, y])
            imgx = int(imgx)
            imgy = int(imgy)
            # Check to see if this overlaps
            _cnt = np.sum(
                np.where(
                    mask[imgx - 15 : imgx + 15, imgy - 15 : imgy + 15], 1, 0
                )
            )
            if _cnt > 5:
                continue
            mask[imgx - 15 : imgx + 15, imgy - 15 : imgy + 15] = True
            # Plot bars
            if stdata.get("sknt") is not None and stdata["sknt"] > 1:
                (u, v) = meteorology.uv(
                    speed(stdata.get("sknt", 0), "KT"),
                    direction(stdata.get("drct", 0), "DEG"),
                )
                if u is not None and v is not None:
                    self.ax.barbs(x, y, u.value("KT"), v.value("KT"), zorder=1)

            # Sky Coverage
            skycoverage = stdata.get("coverage")
            if (
                skycoverage is not None
                and skycoverage >= 0
                and skycoverage <= 100
            ):
                w = Wedge(
                    (x, y), circlesz, 0, 360, ec="k", fc="white", zorder=2
                )
                self.ax.add_artist(w)
                w = Wedge(
                    (x, y),
                    circlesz,
                    0,
                    360.0 * skycoverage / 100.0,
                    ec="k",
                    fc="k",
                    zorder=3,
                )
                self.ax.add_artist(w)

            # Temperature
            val = stdata.get("tmpf")
            if val is not None:
                (offx, offy, ha, va) = offsets[1]
                self.ax.annotate(
                    stdata.get("tmpf_format", "%.0f") % (val,),
                    xy=(x, y),
                    ha=ha,
                    va=va,
                    xytext=(offx, offy),
                    color=stdata.get("tmpf_color", "r"),
                    textcoords="offset points",
                    zorder=Z_OVERLAY + 2,
                    clip_on=True,
                    fontsize=kwargs.get("fontsize", 8),
                )
            # Dew Point
            val = stdata.get("dwpf")
            if val is not None:
                (offx, offy, ha, va) = offsets[7]
                self.ax.annotate(
                    stdata.get("dwpf_format", "%.0f") % (val,),
                    xy=(x, y),
                    ha=ha,
                    va=va,
                    xytext=(offx, offy),
                    color=stdata.get("dwpf_color", "b"),
                    textcoords="offset points",
                    zorder=Z_OVERLAY + 2,
                    clip_on=True,
                    fontsize=kwargs.get("fontsize", 8),
                )
            # Plot identifier
            val = stdata.get("id")
            if val is not None:
                (offx, offy, ha, va) = (
                    offsets[6] if skycoverage is not None else offsets[5]
                )
                self.ax.annotate(
                    "%s" % (val,),
                    xy=(x, y),
                    ha=ha,
                    va=va,
                    xytext=(offx, offy),
                    color=stdata.get("id_color", "tan"),
                    textcoords="offset points",
                    zorder=Z_OVERLAY + 2,
                    clip_on=True,
                    fontsize=kwargs.get("fontsize", 8),
                )

    def plot_values(
        self,
        lons,
        lats,
        vals,
        fmt="%s",
        valmask=None,
        color="#000000",
        textsize=14,
        labels=None,
        labeltextsize=10,
        labelcolor="#000000",
        showmarker=False,
        labelbuffer=25,
        outlinecolor="#FFFFFF",
        zorder=None,
        **kwargs
    ):
        """Plot values onto the map

        Args:
          lons (list): longitude values to use for placing `vals`
          lats (list): latitude values to use for placing `vals`
          vals (list): actual values to place on the map
          fmt (str, optional): Format specification to use for representing the
            values. For example, the default is '%s'.
          valmask (list, optional): Boolean list to use as masking of the
            `vals` while adding to the map.
          color (str, list, optional): Color to use while plotting the `vals`.
            This can be a list to specify each color to use with each value.
          textsize (str, optional): Font size to draw text.
            labels (list, optional): Optional list of labels to place below the
            plotting of `vals`
          labeltextsize (int, optional): Size of the label text
          labelcolor (str, optional): Color to use for drawing labels
          showmarker (bool, optional): Place a marker on the map for the label
          labelbuffer (int): pixel buffer around labels, a value of `0`
            disables the label culling logic.
          outlinecolor (color): color to use for text outlines
          zorder (int or list, optional): zorder to use for plotting.
          textoutlinewidth (int): width of the font outline, default 3.
        """
        if valmask is None:
            valmask = [True] * len(lons)
        if labels is None:
            labels = [""] * len(lons)
        if zorder is None:
            zorder = [Z_OVERLAY + 2] * len(lons)
        if isinstance(color, str):
            color = [color] * len(lons)
        bbox = self.fig.get_window_extent().transformed(
            self.fig.dpi_scale_trans.inverted()
        )
        axbbox = self.ax.get_window_extent().transformed(
            self.fig.dpi_scale_trans.inverted()
        )
        axx0 = axbbox.x0 * self.fig.dpi
        axx1 = (axbbox.x0 + axbbox.width) * self.fig.dpi
        axy0 = axbbox.y0 * self.fig.dpi
        axy1 = (axbbox.y0 + axbbox.height) * self.fig.dpi
        figwidth = bbox.width * self.fig.dpi
        figheight = bbox.height * self.fig.dpi
        if self.textmask is None:
            # zorder is used to track plotting priorities
            self.textmask = np.zeros((int(figwidth), int(figheight)), np.int8)
        thisax = self.ax
        # Create a fake label, to test out our scaling
        t0 = self.fig.text(
            0.5,
            0.5,
            "ABCDEFGHIJ",
            transform=thisax.transAxes,
            color="None",
            size=textsize,
        )
        bbox = t0.get_window_extent(self.fig.canvas.get_renderer())
        xpixels_per_char = bbox.width / 10.0
        ypixels = bbox.height
        for o, a, v, m, c, label, z in zip(
            lons, lats, vals, valmask, color, labels, zorder
        ):
            if not m:
                continue

            ha = "center"
            mystr = fmt % (v,)
            max_mystr_len = max([len(s) for s in mystr.split("\n")])
            mystr_lines = len(mystr.split("\n"))
            # compute the pixel coordinate of this data point
            (x, y) = thisax.projection.transform_point(o, a, ccrs.Geodetic())
            (imgx, imgy) = thisax.transData.transform([x, y])
            imgx0 = int(imgx - (max_mystr_len * xpixels_per_char / 2.0))
            if imgx0 < axx0:
                ha = "left"
                imgx0 = imgx
            imgx1 = imgx0 + max_mystr_len * xpixels_per_char
            if imgx1 > axx1:
                imgx1 = imgx
                imgx0 = imgx1 - max_mystr_len * xpixels_per_char
                ha = "right"
            # Now we buffer
            imgx0 = max([0, imgx0 - labelbuffer])
            imgx1 = min(
                [
                    figwidth,
                    (
                        imgx0
                        + 2 * labelbuffer
                        + max_mystr_len * xpixels_per_char
                    ),
                ]
            )
            imgy0 = max([0, int(imgy) - labelbuffer * 0.75])
            imgy1 = min(
                [
                    figheight,
                    (imgy0 + mystr_lines * ypixels + 2 * labelbuffer * 0.75),
                ]
            )
            overlap = (
                self.textmask[int(imgx0) : int(imgx1), int(imgy0) : int(imgy1)]
                >= z
            )
            _cnt = np.sum(overlap)
            # If we have more than 15 pixels of overlap, don't plot this!
            if _cnt > 15 and labelbuffer > 0:
                if self.debug:
                    LOG.info(
                        "culling |%s| due to overlap, %s", repr(mystr), _cnt
                    )
                continue
            if self.debug:
                rec = plt.Rectangle(
                    [imgx0, imgy0],
                    (imgx1 - imgx0),
                    (imgy1 - imgy0),
                    facecolor="None",
                    edgecolor="r",
                )
                self.fig.patches.append(rec)
            # Useful for debugging this algo
            if self.debug:
                LOG.info(
                    (
                        "label: %s imgx: %s/%s-%s imgy: %s/%s-%s "
                        "x:%s-%s y:%s-%s _cnt:%s"
                    ),
                    repr(mystr),
                    imgx,
                    axx0,
                    axx1,
                    imgy,
                    axy0,
                    axy1,
                    imgx0,
                    imgx1,
                    imgy0,
                    imgy1,
                    _cnt,
                )
            self.textmask[
                int(imgx0) : int(imgx1), int(imgy0) : int(imgy1)
            ] = np.where(
                self.textmask[int(imgx0) : int(imgx1), int(imgy0) : int(imgy1)]
                < z,
                z,
                self.textmask[
                    int(imgx0) : int(imgx1), int(imgy0) : int(imgy1)
                ],
            )
            t0 = thisax.text(
                o,
                a,
                mystr,
                color=c,
                size=textsize,
                zorder=z,
                va="center" if not showmarker else "bottom",
                ha=ha,
                transform=ccrs.PlateCarree(),
            )
            bbox = t0.get_window_extent(self.fig.canvas.get_renderer())
            if self.debug:
                rec = plt.Rectangle(
                    [bbox.x0, bbox.y0],
                    bbox.width,
                    bbox.height,
                    facecolor="None",
                    edgecolor="k",
                )
                self.fig.patches.append(rec)
            if showmarker:
                thisax.scatter(
                    o,
                    a,
                    marker="+",
                    zorder=z,
                    color="k",
                    transform=ccrs.PlateCarree(),
                )
            t0.set_clip_on(True)
            t0.set_path_effects(
                [
                    PathEffects.Stroke(
                        linewidth=kwargs.get("textoutlinewidth", 3),
                        foreground=outlinecolor,
                    ),
                    PathEffects.Normal(),
                ]
            )

            if label and label != "":
                thisax.annotate(
                    "%s" % (label,),
                    xy=(x, y),
                    ha="center",
                    va="top",
                    xytext=(0, 0 - textsize / 2),
                    color=labelcolor,
                    textcoords="offset points",
                    zorder=z - 1,
                    clip_on=True,
                    fontsize=labeltextsize,
                )

    def scatter(self, lons, lats, vals, clevs, **kwargs):
        """Draw points on the map

        Args:
          lons (list): longitude values
          lats (list): latitude values
          vals (list): Data values for the points to use for colormapping
          clevs (list): Levels to use for ramp
          **kwargs: additional options
        """
        cmap = stretch_cmap(
            kwargs.get("cmap"), clevs, extend=kwargs.get("extend")
        )
        norm = mpcolors.BoundaryNorm(clevs, cmap.N)

        colors = cmap(norm(vals))
        self.ax.scatter(
            lons,
            lats,
            c=colors,
            edgecolors=colors,
            transform=ccrs.PlateCarree(),
            zorder=Z_OVERLAY,
        )
        kwargs.pop("cmap", None)
        self.draw_colorbar(clevs, cmap, norm, **kwargs)

    def hexbin(self, lons, lats, vals, clevs, **kwargs):
        """ hexbin wrapper """
        cmap = stretch_cmap(
            kwargs.get("cmap"), clevs, extend=kwargs.get("extend")
        )
        norm = mpcolors.BoundaryNorm(clevs, cmap.N)

        points = self.ax.projection.transform_points(
            ccrs.PlateCarree(), lons, lats
        )
        _hex = self.ax.hexbin(
            points[:, 0],
            points[:, 1],
            C=vals,
            norm=norm,
            cmap=cmap,
            zorder=Z_FILL,
        )
        kwargs.pop("cmap", None)
        self.draw_colorbar(clevs, cmap, norm, **kwargs)
        return _hex

    def pcolormesh(self, lons, lats, vals, clevs, **kwargs):
        """Opinionated mpl.pcolormesh wrapper.

        If you supply a lons in the same shape of the vals, this method will
        tack on an extra row and column to make matplotlib happy. If you do
        not want this, then pass your own lons + lats that is 1 column and 1
        row greater than vals.
        """
        cmap = stretch_cmap(
            kwargs.get("cmap"), clevs, extend=kwargs.get("extend")
        )
        norm = mpcolors.BoundaryNorm(clevs, cmap.N)
        if lons.shape == vals.shape:
            # Tack on some extra
            new = lons[-1, :] * 2 - lons[-2, :]
            lons = np.r_[lons, [new]]
            new = lats[-1, :] * 2 - lats[-2, :]
            lats = np.r_[lats, [new]]
            new = lons[:, -1] * 2 - lons[:, -2]
            lons = np.c_[lons, new]
            new = lats[:, -1] * 2 - lats[:, -2]
            lats = np.c_[lats, new]

        res = self.ax.pcolormesh(
            lons,
            lats,
            vals,
            norm=norm,
            cmap=cmap,
            zorder=Z_FILL,
            transform=ccrs.PlateCarree(),
        )

        if kwargs.get("clip_on", True):
            self.draw_mask()
        kwargs.pop("cmap", None)
        self.draw_colorbar(clevs, cmap, norm, **kwargs)
        return res

    def draw_mask(self, sector=None):
        """Draw a mask on the main axes.

        If `sector` is not provided, this attempts to intelligently to the
        masking the user wants.

        Args:
          sector (str,optional): Hard code what type of sector masking should
            happen.
        """
        sector = self.sector if sector is None else sector
        # can't mask what we don't have a polygon bounds for.
        if sector not in (
            "iailin",
            "midwest",
            "conus",
            "state",
            "iowawfo",
            "cwa",
        ):
            return
        # in lon,lat
        if sector == "state":
            s = load_pickle_geo("us_states.pickle")
            mask_outside_geom(self.ax, s[self.state]["geom"])
            return
        if sector == "cwa":
            s = load_pickle_geo("cwa.pickle")
            mask_outside_geom(self.ax, s[self.cwa]["geom"])
            return
        if sector == "iowawfo":
            s = load_pickle_geo("iowawfo.pickle")
            geo = s["iowawfo"]["geom"]
            ccw = np.asarray(geo.exterior)[::-1]
        else:
            ccw = load_bounds("%s_ccw" % (sector,))
        # in map coords
        points = self.ax.projection.transform_points(
            ccrs.Geodetic(), ccw[:, 0], ccw[:, 1]
        )
        mask_outside_polygon(points[:, :2], ax=self.ax)

    def contourf(self, lons, lats, vals, clevs, **kwargs):
        """Contourf

        Args:
          ilabel (boolean,optional): Should we label contours
          iline (boolean,optional): should we draw contour lines

        Returns:
          vals (np.array): The values used for plotting, maybe after gridding
        """
        if isinstance(lons, list):
            lons = np.array(lons)
            lats = np.array(lats)
            vals = np.array(vals)
        if np.array(vals).ndim == 1:
            # We need to grid, get current plot bounds in display proj
            # Careful here as a rotated projection may have maxes not in ul
            xbnds = self.ax.get_xlim()
            ybnds = self.ax.get_ylim()
            ll = ccrs.Geodetic().transform_point(
                xbnds[0], ybnds[0], self.ax.projection
            )
            ul = ccrs.Geodetic().transform_point(
                xbnds[0], ybnds[1], self.ax.projection
            )
            ur = ccrs.Geodetic().transform_point(
                xbnds[1], ybnds[1], self.ax.projection
            )
            lr = ccrs.Geodetic().transform_point(
                xbnds[1], ybnds[0], self.ax.projection
            )
            xi = np.linspace(min(ll[0], ul[0]), max(lr[0], ur[0]), 100)
            yi = np.linspace(min(ll[1], ul[1]), max(ul[1], ur[1]), 100)
            xi, yi = np.meshgrid(xi, yi)
            nn = NearestNDInterpolator((lons, lats), vals)
            vals = nn(xi, yi)
            lons = xi
            lats = yi
            window = np.ones((6, 6))
            vals = convolve2d(
                vals, window / window.sum(), mode="same", boundary="symm"
            )
        if lons.ndim == 1:
            lons, lats = np.meshgrid(lons, lats)

        cmap = stretch_cmap(
            kwargs.get("cmap"), clevs, extend=kwargs.get("extend")
        )
        norm = mpcolors.BoundaryNorm(clevs, cmap.N)
        # vals = maskoceans(lons, lats, vals, resolution='h')
        self.ax.contourf(
            lons,
            lats,
            vals,
            clevs,
            cmap=cmap,
            norm=norm,
            zorder=Z_FILL,
            extend="both",
            transform=ccrs.PlateCarree(),
        )
        if kwargs.get("iline", True):
            csl = self.ax.contour(
                lons,
                lats,
                vals,
                clevs,
                colors="w",
                zorder=Z_FILL_LABEL,
                transform=ccrs.PlateCarree(),
            )
            if kwargs.get("ilabel", False):
                self.ax.clabel(
                    csl,
                    fmt=kwargs.get("labelfmt", "%.0f"),
                    colors="k",
                    fontsize=14,
                )
        if kwargs.get("clip_on", True):
            self.draw_mask()
        kwargs.pop("cmap", None)
        self.draw_colorbar(clevs, cmap, norm, **kwargs)
        return vals

    def fill_climdiv(self, data, **kwargs):
        """Fill climate divisions using provided data dictionary

        Args:
          data (dict): A dictionary of climate division IDs and values
          bins (optional[list]): a list of values for classification
          lblformat (optional[str]): a format specifier to use
          **kwargs (optional): other values
        """
        clidict = load_pickle_geo("climdiv.pickle")
        polygon_fill(self, clidict, data, **kwargs)

    def fill_ugcs(self, data, bins=None, color=None, **kwargs):
        """Overlay filled UGC geometries

        The logic for plotting is a bit tricky due to fire zones overlapping
        forecast zones.  In general, provide the zone code if you want it to
        display on top.  Otherwise, I attempt to place forecast zones overtop
        fire weather zones.

        Args:
          data(dict): A dictionary of 6 char UGC code keys and values
          bins(list, optional): Bins to use for cloropleth, default 0:101:10
          color(dict, optional): Hard code what each UGC should display as
            for color.
          nocbar (bool, optional): Should a color bar be generated, default is
            `True`.
          plotmissing(bool, optional): Should missing UGC data be plotted?
          labels(dict, optional): UGC indexed dictionary to use for labeling.
          missingval(str, optional): value to use when labelling UGCs with
            missing values, defaults to '-'.
        """
        if bins is None:
            bins = np.arange(0, 101, 10)
        cmap = stretch_cmap(
            kwargs.get("cmap"), bins, extend=kwargs.get("extend")
        )
        norm = mpcolors.BoundaryNorm(bins, cmap.N)
        # Figure out if we have zones or counties/parishes
        counties = True
        for key in data:
            if key[2] == "Z":
                counties = False
            break
        ugcs = load_pickle_geo(
            "ugcs_county.pickle" if counties else "ugcs_zone.pickle"
        )
        if color is None:
            color = dict()
        filter_func = true_filter
        if self.sector == "state":
            filter_func = state_filter
        elif self.sector == "cwa":
            filter_func = cwa_filter
        ilabel = kwargs.get("ilabel", False)
        plotmissing = kwargs.get("plotmissing", True)
        labels = kwargs.get("labels", dict())
        to_label = {"x": [], "y": [], "vals": [], "zorder": []}
        for ugc in ugcs:
            ugcdict = ugcs[ugc]
            if not filter_func(self, ugc, ugcdict):
                continue
            if data.get(ugc) is None:
                if not plotmissing:
                    continue
                # Holy cow, it appears values above 300 are always firewx,
                # so lets ignore these when we have no data!
                if not counties and int(ugc[3:]) >= 300:
                    continue
                c = "white"
                val = kwargs.get("missingval", "-")
                z = Z_OVERLAY
            else:
                val = data[ugc]
                c = color.get(ugc, cmap(norm([val]))[0])
                z = Z_OVERLAY2
            for polyi, polygon in enumerate(ugcdict.get("geom", [])):
                if polygon.exterior is None:
                    continue
                if z in [Z_OVERLAY, Z_OVERLAY2]:
                    arr = np.asarray(polygon.exterior)
                    # Figure out which ax this goes with
                    ax = self.ax
                    if self.sector == "nws":
                        if ugc.startswith("PR"):
                            ax = self.pr_ax
                        elif ugc.startswith("AK"):
                            ax = self.ak_ax
                        elif ugc.startswith("HI"):
                            ax = self.hi_ax
                    points = ax.projection.transform_points(
                        ccrs.Geodetic(), arr[:, 0], arr[:, 1]
                    )
                    ax.add_patch(
                        Polygon(points[:, :2], fc=c, ec="k", zorder=z, lw=0.1)
                    )
                if polyi == 0 and ilabel:
                    # prefer our stored centroid vs calculated one
                    mx = ugcdict.get("lon", polygon.centroid.x)
                    my = ugcdict.get("lat", polygon.centroid.y)
                    to_label["x"].append(mx)
                    to_label["y"].append(my)
                    to_label["vals"].append("%s" % (labels.get(ugc, val),))
                    to_label["zorder"].append(z + 1)
        if to_label:
            self.plot_values(
                to_label["x"],
                to_label["y"],
                to_label["vals"],
                zorder=to_label["zorder"],
                labelbuffer=kwargs.get("labelbuffer", 1),
                textsize=12,
                textoutlinewidth=2,
            )
        if "cmap" in kwargs:
            del kwargs["cmap"]
        if not kwargs.get("nocbar", False):
            self.draw_colorbar(bins, cmap, norm, **kwargs)

    def fill_states(self, data, **kwargs):
        """Add overlay of filled state polygons"""
        states = load_pickle_geo("us_states.pickle")
        polygon_fill(self, states, data, **kwargs)

    def draw_cwas(self, color="k", **kwargs):
        """Overlay CWA Borders

        Draw the CWA border lines on the map.

        Args:
          color(str): The color to draw the CWA borders with
          kwargs(dict, optional): Parameters passed to matplotlib for plotting
        """
        kwargs["edgecolor"] = color
        cwas = load_pickle_geo("cwa.pickle")
        for _a in self.axes:
            _a.add_geometries(
                [val["geom"] for key, val in cwas.items()],
                crs=ccrs.PlateCarree(),
                zorder=Z_POLITICAL,
                facecolor="None",
                **kwargs
            )

    def fill_cwas(self, data, **kwargs):
        """Add overlay of filled polygons for NWS Forecast Offices.

        Method adds a colorized overlay of NWS Forecast Offices based on a
        data dictionary of values provided. This method also places a color
        bar on the image.

        Args:
          data (dict): Dictionary of values with keys representing the 3 char
            or 4 char idenitifer for the WFO.  This assumes the 3 char sites
            are the K ones.
          labels (dict, optional): Optional dictionary that follows the
            ``data`` attribute, but hard codes what should be plotted as a
            label.
          bins (list, optional): List of increasing values to use as bins to
            determine color levels.
          lblformat (str, optional): Format string to use to place labels.
          cmap (matplotlib.cmap, optional): Colormap to use with ``bins``

        """
        cwas = load_pickle_geo("cwa.pickle")
        # Painfull.  San Juan's WFO identifier is SJU, but VTEC uses JSJ, our
        # plotting here uses SJU
        if "JSJ" in data:
            data["SJU"] = data["JSJ"]
        polygon_fill(self, cwas, data, **kwargs)

    def drawcities(self, **kwargs):
        """Overlay some cities

        Args:
          minarea (int,optional): Minimum Urban Area size (km2) to plot
          labelbuffer (int): approximate number of pixels to compute overlap
          textsize (int): size of the text
          color (str): color to plot the text with
          outlinecolor (str): color to outline the text with
        """
        df = load_pickle_pd("pd_cities.pickle")
        (west, east, south, north) = self.ax.get_extent(crs=ccrs.PlateCarree())

        minarea = kwargs.get(
            "minarea", 500.0 if self.sector in ["nws", "conus"] else 10.0
        )
        df2 = df[
            (
                (df["lat"] > south)
                & (df["lat"] < north)
                & (df["lon"] > west)
                & (df["lon"] < east)
                & (df["area_km2"] > minarea)
            )
        ]
        # debug option to test an individual point on the plot
        # df2 = df[(df['name'] == 'Sioux City')]
        # hack around a API break
        tsz = kwargs.pop("textsize", 16)
        self.plot_values(
            df2.lon.values,
            df2.lat.values,
            df2.name.values,
            showmarker=True,
            textsize=tsz,
            **kwargs
        )

    def drawcounties(self, color="k"):
        """Draw counties onto the map

        Args:
          color (color,optional): line color to use
        """
        ugcdict = load_pickle_geo("ugcs_county.pickle")
        polys = []
        for ugc in ugcdict:
            for polygon in ugcdict[ugc].get("geom", []):
                polys.append(polygon)
        self.ax.add_geometries(
            polys,
            crs=ccrs.PlateCarree(),
            facecolor="None",
            edgecolor=color,
            lw=0.4,
            zorder=Z_POLITICAL,
        )

    def iemlogo(self):
        """Place the IEM Logo"""
        draw_logo(self.fig, "logo.png")

    def postprocess(self, **kwargs):
        """Postprocessing.

        Args:
          filename (str): file to save output to.
          web (bool): Write result to sys.stdout, default False.
          memcache (obj): write image to memcache
          memcachekey (str): key to use when writing to memcache.
          memcacheexpire (int): how long should we persist in memcache,
            default is 300.
          pqstr (str): Do pqinsert with the following LDM product name.
        """
        ram = BytesIO()
        self.fig.savefig(ram, format="png")
        ram.seek(0)
        im = Image.open(ram)
        im2 = im.convert("RGB").convert("P", palette=Image.ADAPTIVE)
        if kwargs.get("memcache") and kwargs.get("memcachekey"):
            ram = BytesIO()
            im2.save(ram, format="png")
            ram.seek(0)
            r = ram.read()
            kwargs.get("memcache").set(
                kwargs.get("memcachekey"),
                r,
                time=kwargs.get("memcacheexpire", 300),
            )
        if kwargs.get("web", False):
            ssw("Content-Type: image/png\n\n")
            im2.save(getattr(sys.stdout, "buffer", sys.stdout), format="png")
            return
        tmpfd = tempfile.NamedTemporaryFile(delete=False)
        im2.save(tmpfd, format="PNG")
        tmpfd.close()
        if kwargs.get("pqstr") is not None:
            subprocess.call(
                "pqinsert -p '%s' %s" % (kwargs.get("pqstr"), tmpfd.name),
                shell=True,
            )
        if kwargs.get("filename") is not None:
            shutil.copyfile(tmpfd.name, kwargs.get("filename"))
        os.unlink(tmpfd.name)


def windrose(*args, **kwargs):
    """Depreciated."""
    warnings.warn(
        "windrose() is depreciated, use pyiem.windrose_utils!",
        DeprecationWarning,
    )
    import pyiem.windrose_utils as wru

    return wru.windrose(*args, **kwargs)



"""Test our DSM Parsing."""
# pylint: disable=redefined-outer-name
import datetime

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import pytest
from pyiem.util import utc, get_test_file
from pyiem.nws.products.dsm import process, parser
from pyiem.util import get_dbconn


@pytest.fixture
def dbcursor():
    """Get a database cursor for testing."""
    cursor = get_dbconn("iem").cursor()
    # Create fake station, so we can create fake entry in summary
    # and current tables
    cursor.execute(
        """
        INSERT into stations(id, network, iemid, tzname)
        VALUES ('HKS', 'ZZ_ASOS', -100, 'UTC')
    """
    )
    cursor.execute(
        """
        INSERT into summary_2015(iemid, day) VALUES
        (-100, '2015-11-26')
    """
    )
    return cursor


@pytest.mark.parametrize("month", range(1, 13))
def test_simple(month):
    """Can we walk before we run."""
    text = (
        "KCVG DS 24/%02i 590353/ 312359// 53/ 48/"
        "/9470621/T/T/00/00/00/00/00/00/"
        "00/00/00/00/00/00/00/00/00/00/00/00/00/00/00/00/00/225/26381759/"
        "26500949="
    ) % (month,)
    tzprovider = {"KCVG": ZoneInfo("America/New_York")}
    dsm = process(text)
    dsm.compute_times(utc(2019, month, 25))
    dsm.tzlocalize(tzprovider["KCVG"])
    assert dsm.date == datetime.date(2019, month, 24)
    assert dsm.station == "KCVG"
    assert dsm.time_sped_max == utc(2019, month, 24, 22, 59)


def test_collective(dbcursor):
    """Can we parse a collective."""
    prod = parser(get_test_file("DSM/DSM.txt"), utc(2015, 11, 26))
    assert not prod.warnings
    assert len(prod.data) == 23

    res = prod.sql(dbcursor)
    # first database insert should work from above
    assert res[0]


def test_200824_refail():
    """Test a RE parse failure."""
    text = (
        "KHEI DS 1500 23/08 M/M// 86/M//7061455/00/00/00/00/00/00/00/00/00/00/"
        "00/00/00/00/00/00/-187/-51/-/-/-/-/-/-/-/-/06231121/05311119="
    )
    dsm = process(text)
    assert dsm is not None


def test_190225_regress():
    """Parse something that failed RE."""
    text = (
        "KCUT DS 1100 25/02 061059/-040848//06/-03//0110442/"
        "T/00/00/00/00/00/00/00/00/00/00/T/-/-/-/-/-/-/-/-/-/-/-/-/-/-/"
        "08070124/10090103/18"
    )
    dsm = process(text)
    assert dsm is not None
    dsm.compute_times(utc(2019, 2, 25))
    assert dsm.station == "KCUT"


def test_190225_badtime():
    """This should not trip us up."""
    text = (
        "KMMV DS 1500 05/02 361459/ 250741// 36/ 25//9740006/01/T/T/01/T/00/"
        "00/00/00/00/00/00/00/00/00/00/-/-/-/-/-/-/-/-/-/-/04090030/211750614/"
        "1="
    )
    dsm = process(text)
    assert dsm is not None
    dsm.compute_times(utc(2019, 2, 5))
    assert dsm.station == "KMMV"



"""test lsr."""

from pyiem.nws.products.lsr import _mylowercase
from pyiem.nws.products import lsr


def test_mag_string():
    """ Magnitude string for various events """
    lr = lsr.LSR()
    lr.typetext = "TSTM WND GST"
    lr.magnitude_units = "MPH"
    lr.magnitude_f = 59
    lr.magnitude_qualifier = "M"
    assert lr.mag_string() == "TSTM WND GST of M59 MPH"


def test_get_dbtype():
    """ See what we get for a given LSR typetext """
    lr = lsr.LSR()
    lr.typetext = "TORNADO"
    assert lr.get_dbtype() == "T"


def test_lowercase():
    """ Make sure we can properly convert cities to mixed case """
    assert _mylowercase("1 N AMES") == "1 N Ames"
    assert _mylowercase("1 NNW AMES") == "1 NNW Ames"



"""Plotting"""
# pylint: disable=unused-import

from pyiem.plot.colormaps import *
from pyiem.plot.geoplot import *
from pyiem.plot.calendarplot import *
from pyiem.plot.util import centered_bins, pretty_bins



"""Can we process the SAW"""
# pylint: disable=redefined-outer-name

import psycopg2.extras
import pytest
from pyiem.nws.products.saw import parser as sawparser
from pyiem.util import get_dbconn, utc, get_test_file


@pytest.fixture
def dbcursor():
    """Get database cursor."""
    return get_dbconn("postgis").cursor(
        cursor_factory=psycopg2.extras.RealDictCursor
    )


def test_181231_linkisok():
    """The plain text tweet should have a space."""
    utcnow = utc(2014, 3, 10, 3, 29)
    utcnow = utcnow.replace(microsecond=100)
    prod = sawparser(get_test_file("SAW/SAW3.txt"), utcnow=utcnow)
    assert prod.ets.microsecond == 0
    jmsgs = prod.get_jabbers("")
    ans = (
        "SPC issues Severe Thunderstorm Watch 503 till 9:00Z "
        "https://www.spc.noaa.gov/products/watch/2014/ww0503.html"
    )
    assert jmsgs[0][0] == ans
    assert jmsgs[0][2]["channels"] == "SPC,SPC.SVRWATCH"


def test_replacement(dbcursor):
    """Can we do replacements?"""
    utcnow = utc(2017, 8, 21, 9, 17)
    prod = sawparser(get_test_file("SAW/SAW-replaces.txt"), utcnow=utcnow)
    prod.sql(dbcursor)
    jmsgs = prod.get_jabbers("")
    assert len(jmsgs) == 1
    ans = (
        "SPC issues Severe Thunderstorm Watch"
        " 153 till 17:00Z, new watch replaces WW 1 "
        "https://www.spc.noaa.gov/products/watch/2017/ww0153.html"
    )
    assert jmsgs[0][0] == ans
    assert "twitter" in jmsgs[0][2]


def test_saw3():
    """SAW3"""
    utcnow = utc(2014, 3, 10, 3, 29)
    sts = utcnow.replace(hour=3, minute=35)
    ets = utcnow.replace(hour=9, minute=0)
    prod = sawparser(get_test_file("SAW/SAW3.txt"), utcnow=utcnow)
    assert prod.saw == 3
    assert abs(prod.geometry.area - 7.73) < 0.01
    assert prod.ww_num == 503
    assert prod.sts == sts
    assert prod.ets == ets
    assert prod.ww_type == prod.SEVERE_THUNDERSTORM
    assert prod.action == prod.ISSUES


def test_cancelled(dbcursor):
    """SAW-cancelled make sure we can cancel a watch"""
    utcnow = utc(2014, 3, 10, 3, 29)
    prod = sawparser(get_test_file("SAW/SAW-cancelled.txt"), utcnow=utcnow)
    assert prod.action == prod.CANCELS
    j = prod.get_jabbers(None)
    ans = (
        "Storm Prediction Center cancels Weather Watch Number 575 "
        "https://www.spc.noaa.gov/products/watch/2014/ww0575.html"
    )
    assert j[0][0] == ans
    prod.sql(dbcursor)
    prod.compute_wfos(dbcursor)



"""Can we process the SCP"""
# pylint: disable=redefined-outer-name

import psycopg2.extras
import pytest
from pyiem.nws.products.scp import parser
from pyiem.util import get_dbconn, utc, get_test_file


@pytest.fixture
def dbcursor():
    """Get database cursor."""
    dbconn = get_dbconn("asos")
    yield dbconn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    dbconn.close()


def test_201026(dbcursor):
    """Basic test."""
    utcnow = utc(2020, 10, 26, 15, 5)
    prod = parser(get_test_file("NESDIS/SCPPR2.txt"), utcnow=utcnow)
    assert len(prod.data) == 8
    inserts = prod.sql(dbcursor)
    assert inserts == 8


def test_jan1():
    """Test that we get a product crossing a year right."""
    utcnow = utc(2020, 1, 1, 0, 30)
    prod = parser(get_test_file("NESDIS/SCPCR1_jan1.txt"), utcnow=utcnow)
    assert prod.data[0].valid == utc(2019, 12, 31, 23, 57)



"""Test DEP"""
import os
import datetime

from pyiem import dep


def get_path(name):
    """helper"""
    basedir = os.path.dirname(__file__)
    return "%s/../data/wepp/%s" % (basedir, name)


def test_ramps():
    """Ramps should be of length 11"""
    for q in dep.RAMPS:
        for val in dep.RAMPS[q]:
            assert len(val) == 9


def test_scenarios():
    """Can we load scenarios?"""
    df = dep.load_scenarios()
    assert not df.empty
    assert 0 in df.index


def test_cli_fname():
    """Do we get the right climate file names?"""
    res = dep.get_cli_fname(-95.5, 42.5, 0)
    assert res == "/i/0/cli/095x042/095.50x042.50.cli"
    res = dep.get_cli_fname(-97.9999, 42.0, 0)
    assert res == "/i/0/cli/098x042/098.00x042.00.cli"


def test_crop():
    """Read a crop file."""
    df = dep.read_crop(get_path("crop.txt"))
    assert len(df.index) > 10
    assert abs(df["lai"].max() - 9.00) < 0.01
    assert abs(df["avg_temp_c"].max() - 24.30) < 0.01


def test_yld():
    """Read a slope file"""
    df = dep.read_yld(get_path("yld.txt"))
    assert len(df.index) == 10
    assert abs(df["yield_kgm2"].max() - 0.93) < 0.01


def test_slp():
    """Read a slope file"""
    slp = dep.read_slp(get_path("slp.txt"))
    assert len(slp) == 5
    assert abs(slp[4]["y"][-1] + 2.91) < 0.01
    assert abs(slp[4]["slopes"][-1] - 0.033) < 0.01


def test_man():
    """Read a management file please"""
    manfile = dep.read_man(get_path("man.txt"))
    assert manfile["nop"] == 5
    assert manfile["nini"] == 2
    assert manfile["nsurf"] == 2
    assert manfile["nwsofe"] == 3
    assert manfile["nrots"] == 1
    assert manfile["nyears"] == 11

    manfile = dep.read_man(get_path("man2.txt"))
    assert manfile["nop"] == 0


def test_ofe():
    """Read an OFE please"""
    df = dep.read_ofe(get_path("ofe.txt"))
    assert abs(df["precip"].max() - 107.56) < 0.01

    df = dep.read_ofe(get_path("ofe2.txt"))
    print(df["sedleave"].sum())
    assert abs(df["sedleave"].sum() - 400257.48) < 0.01


def test_wb():
    """read a WB file please"""
    df = dep.read_wb(get_path("wb.txt"))
    assert abs(df["precip"].max() - 162.04) < 0.01


def test_cli():
    """read a CLI file please"""
    df = dep.read_cli(get_path("cli.txt"))
    assert len(df.index) == 4018


def test_cli_rfactor():
    """read a CLI file please"""
    df = dep.read_cli(get_path("cli.txt"), compute_rfactor=True)
    assert abs(df["rfactor"].max() - 872.63) < 0.01
    assert (df.groupby(df.index.year).sum()["rfactor"].max() - 4276.60) < 0.01


def test_empty():
    """don't error out on an empty ENV"""
    df = dep.read_env(get_path("empty_env.txt"))
    assert df.empty


def test_read():
    """Read a ENV file"""
    df = dep.read_env(get_path("good_env.txt"))
    df2 = df[df["date"] == datetime.datetime(2010, 6, 5)]
    assert len(df2.index) == 1
    row = df2.iloc[0]
    assert row["runoff"] == 86.3


def test_rfactor_empty():
    """Test our R-factor code."""
    res = dep.rfactor([], [])
    assert res == 0


def test_rfactor_one():
    """Test our R-factor code."""
    # 1 inch rain over 1 hour
    res = dep.rfactor([1.0, 2.0], [0.0, 25.4])
    assert abs(res - 170.31) < 0.1


def test_rfactor_english():
    """Test our R-factor code."""
    # 1 inch rain over 1 hour
    res = dep.rfactor([1.0, 2.0], [0.0, 25.4], return_rfactor_metric=False)
    assert abs(res - 2.71) < 0.1



""" Parse SIGMETs """
# Stdlib imports
import re
import datetime
import math

# Third Party
from shapely.geometry import Polygon, Point

# Local stuff
from pyiem.nws.product import TextProduct

O_LINE1 = re.compile(
    "SIGMET (?P<name>[A-Z]*) (?P<num>[0-9]*) "
    "VALID (?P<sts>[0-9]{6})/(?P<ets>[0-9]{6})"
)

O_PAIRS = re.compile(r"(?P<lat>[NS][0-9]{2,4})\s?(?P<lon>[EW][0-9]{3,5})")

CS_RE = re.compile(
    r"""CONVECTIVE\sSIGMET\s(?P<label>[0-9A-Z]+)\s
VALID\sUNTIL\s(?P<hour>[0-2][0-9])(?P<minute>[0-5][0-9])Z\s
(?P<states>[A-Z ]+)\s
(?P<from>FROM)?\s?(?P<locs>[0-9A-Z \-]+?)\s
(?P<dmshg>DMSHG|DVLPG|INTSF)?\s?(?P<geotype>AREA|LINE|ISOL)?\s?
(?P<cutype>EMBD|SEV|SEV\sEMBD|EMBD\sSEV)?\s?TS\s(?P<width>[0-9]+\sNM\sWIDE)?
(?P<diameter>D[0-9]+)?
""",
    re.VERBOSE,
)

FROM_RE = re.compile(
    r"""
(?P<offset>[0-9]+)?
(?P<drct>N|NE|NNE|ENE|E|ESE|SE|SSE|S|SSW|SW|WSW|W|WNW|NW|NNW)?\s?
(?P<loc>[A-Z0-9]{3})
""",
    re.VERBOSE,
)

OL_RE = re.compile(
    r"OUTLOOK\sVALID\s(?P<begin>[0-9]{6})-(?P<end>[0-9]{6})\n", re.VERBOSE
)

AREA_RE = re.compile(
    r"AREA\s(?P<areanum>[0-9]+)\.\.\.FROM\s(?P<locs>[0-9A-Z \-]+)\n",
    re.VERBOSE,
)

LINE_RE = re.compile(
    r"(?P<distance>[0-9]*)NM\s+EITHER\s+SIDE\s+OF\s+LINE\s+", re.VERBOSE
)

CIRCLE_RE = re.compile(r"WI\s+(?P<distance>[0-9]*)NM\s+OF\s+", re.VERBOSE)


dirs = {
    "NNE": 22.5,
    "ENE": 67.5,
    "NE": 45.0,
    "E": 90.0,
    "ESE": 112.5,
    "SSE": 157.5,
    "SE": 135.0,
    "S": 180.0,
    "SSW": 202.5,
    "WSW": 247.5,
    "SW": 225.0,
    "W": 270.0,
    "WNW": 292.5,
    "NW": 315.0,
    "NNW": 337.5,
    "N": 0,
    "": 0,
}

KM_SM = 1.609347


class SIGMET:
    """Data Structure."""

    def __init__(self):
        """ Constructor """
        self.sts = None
        self.ets = None
        self.geom = None
        self.label = None
        self.areatext = ""
        self.centers = []
        self.raw = None


SIGMETException = Exception


def figure_expire(ptime, hour, minute):
    """
    Convert something like 0255Z into a full blown time
    """
    expire = ptime
    if hour < ptime.hour:
        expire += datetime.timedelta(days=1)
    return expire.replace(hour=hour, minute=minute)


def go2lonlat(lon0, lat0, direction, displacement):
    """http://stackoverflow.com/questions/7222382"""
    # Radius of the Earth
    R = 6378.1
    # Bearing is 90 degrees converted to radians.
    brng = math.radians(dirs.get(direction, 0))
    # Distance in km
    d = displacement / KM_SM

    # Current lat point converted to radians
    lat1 = math.radians(lat0)
    # Current long point converted to radians
    lon1 = math.radians(lon0)

    lat2 = math.asin(
        math.sin(lat1) * math.cos(d / R)
        + math.cos(lat1) * math.sin(d / R) * math.cos(brng)
    )

    lon2 = lon1 + math.atan2(
        math.sin(brng) * math.sin(d / R) * math.cos(lat1),
        math.cos(d / R) - math.sin(lat1) * math.sin(lat2),
    )

    lat2 = math.degrees(lat2)
    lon2 = math.degrees(lon2)

    return lon2, lat2


def locs2lonslats(loc_provider, locstr, geotype, widthstr, diameterstr):
    """
    Convert a locstring into a lon lat arrays
    """
    lats = []
    lons = []
    for l in locstr.split("-"):
        s = FROM_RE.search(l)
        if s:
            d = s.groupdict()
            if d["offset"] is not None:
                (lon1, lat1) = go2lonlat(
                    loc_provider[d["loc"]]["lon"],
                    loc_provider[d["loc"]]["lat"],
                    d["drct"],
                    float(d["offset"]),
                )
            else:
                (lon1, lat1) = (
                    loc_provider[d["loc"]]["lon"],
                    loc_provider[d["loc"]]["lat"],
                )
            lats.append(lat1)
            lons.append(lon1)
    if geotype == "ISOL" or diameterstr is not None:
        lats2 = []
        lons2 = []
        diameter = float(diameterstr.replace("D", ""))
        # Approximation
        diameterdeg = diameter / 110.0
        # UR
        lons2.append(lons[0] - diameterdeg)
        lats2.append(lats[0] + diameterdeg)
        # UL
        lons2.append(lons[0] + diameterdeg)
        lats2.append(lats[0] + diameterdeg)
        # LL
        lons2.append(lons[0] + diameterdeg)
        lats2.append(lats[0] - diameterdeg)
        # LR
        lons2.append(lons[0] - diameterdeg)
        lats2.append(lats[0] - diameterdeg)
        lons = lons2
        lats = lats2

    if geotype == "LINE":
        lats2 = []
        lons2 = []
        # Figure out left hand points
        for i in range(0, len(lats) - 1):
            deltax = lons[i + 1] - lons[i]
            deltay = lats[i + 1] - lats[i]
            if deltax == 0:
                deltax = 0.001
            angle = math.atan(deltay / deltax)
            runx = 0.1 * math.cos(angle)
            runy = 0.1 * math.sin(angle)
            # UR
            lons2.append(lons[i] - runy)
            lats2.append(lats[i] + runx)
            # UL
            lons2.append(lons[i + 1] - runy)
            lats2.append(lats[i + 1] + runx)

        for i in range(0, len(lats) - 1):
            deltax = lons[i + 1] - lons[i]
            deltay = lats[i + 1] - lats[i]
            if deltax == 0:
                deltax = 0.001
            angle = math.atan(deltay / deltax)
            runx = 0.1 * math.cos(angle)
            runy = 0.1 * math.sin(angle)
            # LL
            lons2.append(lons[i + 1] + runy)
            lats2.append(lats[i + 1] - runx)
            # LR
            lons2.append(lons[i] + runy)
            lats2.append(lats[i] - runx)

        lons = lons2
        lats = lats2

    return lons, lats


def compute_esol(pts, distance):
    """ Figure out the box points given the two points and the distance """
    newpts = []
    deltax = pts[1][0] - pts[0][0]
    deltay = pts[1][1] - pts[0][1]
    # Compute unit vector
    linedistance = (deltax ** 2 + deltay ** 2) ** 0.5
    deltax = deltax / linedistance
    deltay = deltay / linedistance
    N = distance / 111.0  # approx
    newpts.append([pts[0][0] - N * deltay, pts[0][1] + N * deltax])
    newpts.append([pts[1][0] - N * deltay, pts[1][1] + N * deltax])
    newpts.append([pts[1][0] + N * deltay, pts[1][1] - N * deltax])
    newpts.append([pts[0][0] + N * deltay, pts[0][1] - N * deltax])
    newpts.append([newpts[0][0], newpts[0][1]])

    return newpts


class SIGMETProduct(TextProduct):
    """
    Represents a Storm Prediction Center Mesoscale Convective Discussion
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        self.sigmets = []
        if self.afos in ["SIGC", "SIGW", "SIGE", "SIGAK1", "SIGAK2"]:
            self.process_SIGC()
        elif self.afos[:2] == "WS":
            self.process_WS()
        else:
            self.process_ocean()

    def sql(self, txn):
        """ Do SQL related stuff that is required """
        txn.execute("DELETE from sigmets_current where expire < now()")
        for sigmet in self.sigmets:
            sqlwkt = f"SRID=4326;{sigmet.geom.wkt}"
            for table in ("sigmets_current", "sigmets_archive"):
                sql = f"DELETE from {table} where label = %s and expire = %s"
                args = (sigmet.label, sigmet.ets)
                txn.execute(sql, args)
                sql = (
                    f"INSERT into {table} (sigmet_type, label, issue, "
                    "expire, raw, geom) VALUES ('C',%s, %s, %s, %s, %s)"
                )
                args = (
                    sigmet.label,
                    self.valid,
                    sigmet.ets,
                    sigmet.raw,
                    sqlwkt,
                )
                txn.execute(sql, args)
            # Compute who is impacted by this SIGMET
            txn.execute(
                "SELECT distinct id from cwsu WHERE "
                f"st_overlaps(geomFromEWKT('{sqlwkt}'), geom) or "
                f"st_contains(geomFromEWKT('{sqlwkt}'), geom) "
            )
            for row in txn.fetchall():
                sigmet.centers.append(row["id"])

    def compute_time(self, ddhhmi):
        """ Convert this string into a proper date time """
        day = int(ddhhmi[:2])
        hour = int(ddhhmi[2:4])
        minute = int(ddhhmi[4:6])
        ts = self.valid
        if self.valid.day > 25 and day < 5:  # next month
            ts += datetime.timedelta(days=15)

        return ts.replace(day=day, hour=hour, minute=minute)

    def process_ocean(self):
        """ Process oceananic """
        meat = self.unixtext.replace("\n", " ")
        m = O_LINE1.search(meat)
        d = m.groupdict()
        if d is None:
            raise SIGMETException("Failed to parse O_LINE1: %s" % (meat,))
        s = SIGMET()
        s.label = "%s %s" % (d["name"], d["num"])
        s.sts = self.compute_time(d["sts"])
        s.ets = self.compute_time(d["ets"])
        m = re.findall(O_PAIRS, meat)
        if not m:
            # TODO: resolve what SIGMET cancels are
            if meat.find("CNL SIGMET") > 0 or meat.find("CANCEL SIGMET") > 0:
                return
            raise SIGMETException("Failed to parse 0_PAIRS: %s" % (meat,))
        pts = []
        for pair in m:
            if len(pair[0][1:]) == 2:
                lat = float(pair[0][1:])
            else:
                lat = float(pair[0][1:]) / 100.0
            if pair[0][0] == "S":
                lat = 0 - lat
            if len(pair[1][1:]) == 3:
                lon = float(pair[1][1:])
            else:
                lon = float(pair[1][1:]) / 100.0
            if pair[1][0] == "W":
                lon = 0 - lon
            pts.append((lon, lat))
        m = LINE_RE.search(meat)
        if m is not None:
            d = m.groupdict()
            pts = compute_esol(pts, int(d["distance"]))
        m = CIRCLE_RE.search(meat)
        if m is not None and len(pts) == 1:
            d = m.groupdict()
            # buffer a point, approximate 1 deg as 100 km :/
            s.geom = Point(pts[0]).buffer(float(d["distance"]) * KM_SM / 100.0)
        else:
            s.geom = Polygon(pts)

        s.raw = self.unixtext
        self.sigmets.append(s)

    def process_SIGC(self):
        """ Process this type of SIGMET """
        for section in self.unixtext.split("\n\n"):
            s = CS_RE.search(section.replace("\n", " "))
            if s is None:
                continue
            data = s.groupdict()
            sig = SIGMET()
            sig.label = data["label"]
            sig.areatext = data["states"].replace(" FROM", "")
            sig.ets = figure_expire(
                self.valid, int(data["hour"]), int(data["minute"])
            )
            lons, lats = locs2lonslats(
                self.nwsli_provider,
                data["locs"],
                data["geotype"],
                data["width"],
                data["diameter"],
            )

            if len(lons) <= 2:
                continue
            pts = []
            for lon, lat in zip(lons, lats):
                pts.append((lon, lat))
            if lats[0] != lats[-1] or lons[0] != lons[-1]:
                pts.append((lons[0], lats[0]))
            sig.geom = Polygon(pts)
            sig.raw = section

            self.sigmets.append(sig)

    def process_WS(self):
        """ Process this type of SIGMET """

    def get_jabbers(self, uri, _uri2=None):
        """ Return the Jabber for this sigmet """
        j = []
        for sig in self.sigmets:
            area = " for " + sig.areatext if sig.areatext != "" else ""
            txt = ("%s issues SIGMET %s%s till %s UTC") % (
                self.source,
                sig.label,
                area,
                sig.ets.strftime("%H%M"),
            )
            html = ("<p>%s issues SIGMET %s%s till %s UTC</p>") % (
                self.source,
                sig.label,
                area,
                sig.ets.strftime("%H%M"),
            )
            channels = ["SIGMET.%s" % (i,) for i in sig.centers]
            channels.append("SIGMET.%s" % (self.source[1:],))
            xtra = {"channels": ",".join(channels), "twitter": txt}

            j.append([txt, html, xtra])
        return j


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Helper function """
    return SIGMETProduct(text, utcnow, ugc_provider, nwsli_provider)



"""tests of twistedpg."""

from psycopg2.extras import DictCursor
from pyiem import twistedpg


def test_connect():
    """Does our logic work?"""
    conn = twistedpg.connect(database="postgis", host="iemdb.local")
    cursor = conn.cursor()
    assert isinstance(cursor, DictCursor)



"""Test GINI"""

from pyiem.util import get_test_file
from pyiem.nws import gini


def test_getirramp():
    """Make sure get_irramp works"""
    d = gini.get_ir_ramp()
    assert len(d) == 256


def test_conus():
    """Test processing a national product"""
    sat = gini.GINIZFile(get_test_file("TIGN02", fponly=True))
    assert sat.archive_filename() == "GOES_SUPER_IR_201509281745.png"
    assert sat.awips_grid() == 0
    assert sat.metadata["map_projection"] == 5


def test_gini():
    """ check GINI Processing of Goes East VIS parsing """
    sat = gini.GINIZFile(get_test_file("TIGH05", fponly=True))
    assert sat.archive_filename() == "GOES_HI_WV_201507161745.png"
    assert str(sat) == "TIGH05 KNES 161745 Line Size: 560 Num Lines: 520"
    assert sat.awips_grid() == 208



"""NWS Hydrological Markup Language

Attempt to break up the HML product into atomic data

"""
import re
from datetime import timezone, datetime
import xml.etree.cElementTree as ET

import pandas as pd
import pyiem.nws.product as product
from pyiem.util import LOG

DELIMITER = r"""\<\?xml version="1.0" standalone="yes"\?\>"""


def no999(val):
    """No negative -999 or -9999 please."""
    if val is None or val == "-999" or val == "-9999":
        return None
    return val


def parseUTC(s):
    """Parse an ISO-ish string into UTC timestamp"""
    if s is None:
        return None
    return datetime.strptime(s[:19], "%Y-%m-%dT%H:%M:%S").replace(
        tzinfo=timezone.utc
    )


def parse_xml(token):
    """Attempt to parse the XML into something useful"""
    root = ET.fromstring(token)
    hml = HMLData()
    hml.station = root.attrib["id"]
    hml.stationname = root.attrib.get("name")
    hml.originator = root.attrib.get("originator")
    hml.generationtime = parseUTC(root.attrib["generationtime"])
    for child in root:
        if child.tag not in ["observed", "forecast"]:
            continue
        rows = []
        for datum in child.findall("datum"):
            secondary = datum.find("secondary")
            rows.append(
                dict(
                    name=child.tag,
                    valid=parseUTC(datum.find("valid").text),
                    primary=no999(datum.find("primary").text),
                    secondary=(
                        no999(secondary.text)
                        if secondary is not None
                        else None
                    ),
                )
            )
        mydict = hml.data[child.tag]
        df = pd.DataFrame(rows)
        df["primary"] = pd.to_numeric(df["primary"], errors="coerce")
        df["secondary"] = pd.to_numeric(df["secondary"], errors="coerce")
        mydict["dataframe"] = df
        mydict["issued"] = parseUTC(child.attrib.get("issued"))
        for attr in [
            "primaryName",
            "secondaryName",
            "primaryUnits",
            "secondaryUnits",
        ]:
            mydict[attr] = child.attrib.get(attr)
    return hml


class HMLData:
    """Our data object."""

    def __init__(self):
        """Constructor."""
        self.station = None
        self.stationname = None
        self.originator = None
        self.generationtime = None
        self.data = {
            "observed": dict(
                dataframe=None,
                primaryUnits=None,
                issued=None,
                secondaryUnits=None,
                primaryName=None,
                secondaryName=None,
            ),
            "forecast": dict(
                dataframe=None,
                primaryUnits=None,
                issued=None,
                secondaryUnits=None,
                primaryName=None,
                secondaryName=None,
            ),
        }


class HML(product.TextProduct):
    """ Class for parsing and representing Space Wx Products """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        product.TextProduct.__init__(
            self,
            text,
            utcnow=utcnow,
            ugc_provider=ugc_provider,
            nwsli_provider=nwsli_provider,
        )
        self.data = []
        self.parsing()

    def do_sql_observed(self, cursor, _hml):
        """Process the observed portion of the dataset"""
        ob = _hml.data["observed"]
        if ob["dataframe"] is None:
            return
        df = ob["dataframe"]
        if df.empty:
            return
        for col in ["primary", "secondary"]:
            if ob[col + "Name"] is None:
                continue
            key = "%s[%s]" % (ob[col + "Name"], ob[col + "Units"])
            # Check that we have some non-null data
            df2 = df[pd.notnull(df[col])]
            if df2.empty:
                continue
            minvalid = df2["valid"].min()
            maxvalid = df2["valid"].max()
            cursor.execute(
                """
                DELETE from hml_observed_data WHERE
                station = %s and valid >= %s and valid <= %s and
                key = get_hml_observed_key(%s)
            """,
                (_hml.station, minvalid, maxvalid, key),
            )
            for _, row in df2.iterrows():
                val = row[col]
                if val is None:
                    continue
                cursor.execute(
                    (
                        "INSERT into hml_observed_data "
                        "(station, valid, key, value) "
                        "VALUES (%s, %s, get_hml_observed_key(%s), %s) "
                        "RETURNING key"
                    ),
                    (_hml.station, row["valid"], key, val),
                )
                if cursor.fetchone()[0] is not None:
                    continue
                # Delete the bad row
                cursor.execute(
                    "DELETE from hml_observed_data WHERE station = %s and "
                    "valid = %s and key is null",
                    (_hml.station, row["valid"]),
                )
                # Need to create a new unit!
                cursor.execute(
                    "INSERT into hml_observed_keys(id, label) VALUES ("
                    "(SELECT coalesce(max(id) + 1, 0) from hml_observed_keys),"
                    "%s) RETURNING id",
                    (key,),
                )
                LOG.info("Created key %s for %s", cursor.fetchone()[0], key)

    def do_sql_forecast(self, cursor, _hml):
        """Process the forecast portion of the dataset"""
        fx = _hml.data["forecast"]
        df = fx["dataframe"]
        if df is None:
            return
        if df.empty:
            return
        # Get an id
        cursor.execute(
            """
        INSERT into hml_forecast(station, generationtime, originator,
        product_id, primaryname, secondaryname, primaryunits,
        secondaryunits, issued, forecast_sts, forecast_ets)
        VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
        RETURNING id
        """,
            (
                _hml.station,
                _hml.generationtime,
                _hml.originator,
                self.get_product_id(),
                fx["primaryName"],
                fx["secondaryName"],
                fx["primaryUnits"],
                fx["secondaryUnits"],
                fx["issued"],
                df["valid"].min(),
                df["valid"].max(),
            ),
        )
        fid = cursor.fetchone()[0]
        # Table partitioning is done by issued time
        table = "hml_forecast_data_%s" % (fx["issued"].year,)
        for _, row in fx["dataframe"].iterrows():
            cursor.execute(
                f"INSERT into {table} (hml_forecast_id, valid, primary_value, "
                "secondary_value) VALUES (%s, %s, %s, %s)",
                (fid, row["valid"], row["primary"], row["secondary"]),
            )

    def sql(self, cursor):
        """Persist this information to the database"""
        for _hml in self.data:
            self.do_sql_forecast(cursor, _hml)
            self.do_sql_observed(cursor, _hml)

    def parsing(self):
        """Attempt to parse out what we have found"""
        tokens = re.split(DELIMITER, self.unixtext)
        for token in tokens:
            if token.find("</site>") == -1:
                continue
            content = token.strip()
            try:
                self.data.append(parse_xml(content))
            except Exception as exp:
                self.warnings.append(
                    ("Parsing %s resulted in %s\n%s")
                    % (self.get_product_id(), exp, content)
                )

    def __str__(self):
        """string representation"""
        s = "HML %s\n" % (self.get_product_id(),)
        for _hml in self.data:
            s += "  + SID: %s generationTime: %s\n" % (
                _hml.station,
                _hml.generationtime,
            )
        return s


def parser(buf, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """Parse a HML NOAAPort product

    This may have multiple xml documents inside.

    Args:
      buf (str): What we want to parse
    """
    return HML(buf, utcnow, ugc_provider, nwsli_provider)



"""Can we make calendar plots, yes we can!"""
import datetime

import pytest
from pyiem.plot import calendar_plot


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_issue101():
    """We like June, it is a good month, don't drop it."""
    sts = datetime.date(2017, 5, 29)
    ets = datetime.date(2017, 9, 30)
    data = dict()
    data[datetime.date(2017, 6, 6)] = {"val": "0606"}
    data[datetime.date(2017, 7, 6)] = {"val": "0506"}
    return calendar_plot(
        sts,
        ets,
        data,
        title="Whiz Bang, Wizzardry",
        subtitle="This is officially unofficial and hacky.",
    )


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_calendar12():
    """See if we can make a calendar plot!"""
    sts = datetime.date(2015, 5, 4)
    ets = datetime.date(2016, 4, 15)
    data = dict()
    data[datetime.date(2015, 6, 6)] = {"val": "0606"}
    data[datetime.date(2015, 5, 6)] = {"val": "0506"}
    return calendar_plot(
        sts,
        ets,
        data,
        title="Whiz Bang, Wizzardry",
        subtitle="This is officially unofficial and hacky.",
    )


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_calendar8():
    """See if we can make a calendar plot!"""
    sts = datetime.date(2015, 5, 4)
    ets = datetime.date(2016, 1, 15)
    data = dict()
    data[datetime.date(2015, 6, 6)] = {"val": "0606"}
    data[datetime.date(2015, 5, 6)] = {"val": "0506"}
    return calendar_plot(
        sts,
        ets,
        data,
        title="Whiz Bang, Wizzardry",
        subtitle="This is officially unofficial and hacky.",
    )


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_calendar4():
    """See if we can make a calendar plot!"""
    sts = datetime.date(2015, 5, 4)
    ets = datetime.date(2015, 8, 15)
    data = dict()
    data[datetime.date(2015, 6, 6)] = {"val": "0606"}
    data[datetime.date(2015, 5, 6)] = {"val": "0506"}
    return calendar_plot(
        sts,
        ets,
        data,
        title="Whiz Bang, Wizzardry",
        subtitle="This is officially unofficial and hacky.\nAnd a second line",
    )


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_calendar2():
    """See if we can make a calendar plot!"""
    sts = datetime.date(2015, 5, 4)
    ets = datetime.date(2015, 6, 15)
    data = dict()
    data[datetime.date(2015, 6, 6)] = {"val": "0606"}
    data[datetime.date(2015, 5, 6)] = {"val": "0506"}
    return calendar_plot(
        sts,
        ets,
        data,
        title=(
            "Whiz Bang, Wizzardry. This is even more text and we "
            "have even more."
        ),
        subtitle="This is officially unofficial and hacky.",
    )


@pytest.mark.mpl_image_compare(tolerance=0.01)
def test_calendar():
    """See if we can make a calendar plot!"""
    sts = datetime.date(2015, 5, 4)
    ets = datetime.date(2015, 5, 15)
    data = dict()
    data[datetime.date(2015, 5, 16)] = {"val": 300, "color": "#ff0000"}
    data[datetime.date(2015, 5, 6)] = {"val": 1, "cellcolor": "#0000ff"}
    return calendar_plot(
        sts,
        ets,
        data,
        title="Whiz Bang, Wizzardry",
        subtitle="This is officially unofficial and hacky.",
        heatmap=True,
    )



"""test NWSLI."""

from pyiem.nws.nwsli import NWSLI


def test_simple():
    """ See if we can generate a proper string from a UGCS """
    nwsli = NWSLI("AMWI4", "Iowa All", ["DMX"], -99, 44)
    assert nwsli.id == "AMWI4"

    assert nwsli.get_name() == "Iowa All"



"""Special Weather Statement"""
import re
import datetime

from shapely.geometry import MultiPolygon
from pyiem import reference
from pyiem.nws.product import TextProduct
from pyiem.nws.ugc import ugcs_to_text

SPECIAL_WX_STATEMENT = re.compile("SPECIAL WEATHER STATEMENT", re.I)
TILL = re.compile(" (TILL|UNTIL|THROUGH) [0-9]{1,2}:?[0-5][0-9]", re.I)


def dedup_headline(headline, ugcs, counties, expire):
    """Try to not be redundant

    Args:
      headline (str): our current headline
      ugcs (list(ugc)): list of ugcs this SPS is for
      counties (str): our current parsed string of counties
      expire (str): our current parsed expiration string

    Returns:
       (str): our new counties, which has been deduped
       (str): our new expire, which has been deduped
    """
    if TILL.search(headline):
        expire = ""
    if ugcs:
        hits = 0
        for ugc in ugcs:
            if ugc.name is None:
                continue
            if headline.upper().find(ugc.name.upper()) > -1:
                hits += 1
        if (float(hits) / float(len(ugcs))) > 0.7:
            counties = ""

    return counties, expire


class SPSProduct(TextProduct):
    """A Special Weather Statement"""

    def sql(self, txn):
        """Do database save in the case of a polygon"""
        for seg in self.segments:
            if not seg.sbw:
                continue
            # Figure out an end time
            ets = self.valid + datetime.timedelta(hours=1)
            if seg.ugcexpire is not None:
                ets = seg.ugcexpire
            giswkt = "SRID=4326;%s" % (MultiPolygon([seg.sbw]).wkt,)
            sql = (
                "INSERT into text_products(product, product_id, geom, "
                "issue, expire, pil) values (%s, %s, %s, %s, %s, %s)"
            )
            myargs = (
                self.unixtext,
                self.get_product_id(),
                giswkt,
                self.valid,
                ets,
                self.afos,
            )
            txn.execute(sql, myargs)

    def _get_channels(self, segment):
        """Returns a list of channels for this SPS."""
        channels = self.get_channels()
        for ugc in segment.ugcs:
            channels.append(f"{self.afos}.{ugc}")
            channels.append(str(ugc))
        return channels

    def get_jabbers(self, uri, _uri2=None):
        """return the standard [[text, html, xtra], ] for jabber"""
        res = []
        xtra = {"product_id": self.get_product_id()}
        for seg in self.segments:
            # Skip any segments that don't have UGC information
            if not seg.ugcs:
                continue
            headline = "[No headline was found in SPS]"
            if seg.headlines:
                headline = (seg.headlines[0]).replace("\n", " ")
            elif SPECIAL_WX_STATEMENT.search(seg.unixtext):
                headline = "Special Weather Statement"
            counties = " for %s" % (ugcs_to_text(seg.ugcs),)
            expire = ""
            if seg.ugcexpire is not None:
                expire = " till %s %s" % (
                    (
                        seg.ugcexpire
                        - datetime.timedelta(
                            hours=reference.offsets.get(self.z, 0)
                        )
                    ).strftime("%-I:%M %p"),
                    self.z,
                )
            counties, expire = dedup_headline(
                headline, seg.ugcs, counties, expire
            )
            xtra["channels"] = self._get_channels(seg)
            mess = ("%s issues %s%s%s %s?pid=%s") % (
                self.source[1:],
                headline,
                counties,
                expire,
                uri,
                xtra["product_id"],
            )
            htmlmess = ("<p>%s issues <a href='%s?pid=%s'>%s</a>%s%s</p>") % (
                self.source[1:],
                uri,
                xtra["product_id"],
                headline,
                counties,
                expire,
            )
            xtra["twitter"] = "%s%s%s %s?pid=%s" % (
                headline,
                counties,
                expire,
                uri,
                xtra["product_id"],
            )
            res.append([mess, htmlmess, xtra])

        return res


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """The SPS Parser"""
    return SPSProduct(
        text, utcnow, ugc_provider=ugc_provider, nwsli_provider=nwsli_provider
    )



"""tests"""
import datetime
import os

import requests
from pyiem import mrms
from pyiem.util import utc

PRODUCT = "PrecipRate"
CENTERS = ["mtarchive", "", "bldr", "cprk"]


def test_nofailback(requests_mock):
    """Test that code bails on old date."""
    valid = utc() - datetime.timedelta(days=20)
    requests_mock.get(
        mrms.get_url("mtarchive", valid, PRODUCT), status_code=404
    )
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is None


def test_failback(requests_mock):
    """Test that we can do option 3."""
    valid = utc() + datetime.timedelta(hours=1)
    requests_mock.get(
        mrms.get_url("mtarchive", valid, PRODUCT), status_code=404
    )
    for center in CENTERS[1:]:
        requests_mock.get(
            mrms.get_url(center, valid, PRODUCT), content=b"\x1f\x8bHello"
        )
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is not None
    os.unlink(fn)


def test_exception(requests_mock):
    """Test what happens when we raise an exception."""
    valid = utc() + datetime.timedelta(hours=1)
    for center in CENTERS:
        requests_mock.get(
            mrms.get_url(center, valid, PRODUCT),
            exc=requests.exceptions.ConnectTimeout,
        )
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is None


def test_existing_file():
    """Test that we return once we already have the file on disk."""
    valid = utc()
    fn = "%s_00.00_%s00.grib2.gz" % (PRODUCT, valid.strftime("%Y%m%d-%H%M"))
    with open(f"/tmp/{fn}", "w") as fh:
        fh.write("Hello")
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is not None
    os.unlink(fn)


def test_fetch_failback(requests_mock):
    """Can we get files that we don't have."""
    # A file from the future suffices
    valid = utc() + datetime.timedelta(hours=1)
    for center in CENTERS:
        requests_mock.get(
            mrms.get_url(center, valid, PRODUCT), status_code=404
        )
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is None


def test_fetch(requests_mock):
    """Can we fetch MRMS files?  Yes we can!"""
    valid = utc()
    requests_mock.get(
        mrms.get_url("mtarchive", valid, PRODUCT), content=b"\x1f\x8bHello"
    )
    fn = mrms.fetch(PRODUCT, valid, tmpdir="/tmp")
    assert fn is not None
    with open(fn, "rb") as fh:
        assert fh.read() == b"\x1f\x8bHello"
    os.unlink(fn)


def test_colorramp():
    """See what we can do with a colorramp"""
    cmap = mrms.make_colorramp()
    assert len(cmap) == 256 * 3


def test_write_worldfile():
    """see if we can write a world file"""
    mrms.write_worldfile("/tmp/bah.wld")
    assert os.path.isfile("/tmp/bah.wld")


def test_get_fn():
    """ Test filename func """
    ts = datetime.datetime(2000, 1, 1, 12, 35)
    fn = mrms.get_fn("p24h", ts, 3)
    expected = (
        "/mnt/a4/data/2000/01/01/mrms/tile3/p24h/p24h.20000101.123500.gz"
    )
    assert fn == expected


def test_reader():
    """Can we read the legacy file """
    fn = ("%s/../data/product_examples/1hrad.20130920.190000.gz") % (
        os.path.dirname(__file__),
    )
    metadata, _ = mrms.reader(fn)
    assert abs(metadata["ul_lat"] - 54.99) < 0.01



"""Multi-RADAR Multi-Sensor (MRMS) Helper Functions

Hopefully useful functions to help with the processing of MRMS data
"""
import struct
from datetime import timezone, datetime
import gzip
import os

import requests
import numpy as np

WEST = -130.0
EAST = -60.0
NORTH = 55.0
SOUTH = 20.0
XAXIS = np.arange(WEST, EAST, 0.01)
YAXIS = np.arange(SOUTH, NORTH, 0.01)


def is_gzipped(text):
    """Check that we have gzipped content"""
    return text[:2] == b"\x1f\x8b"


def get_url(center, valid, product):
    """Return the URL given the provided options."""
    fn = "%s_00.00_%s00.grib2.gz" % (product, valid.strftime("%Y%m%d-%H%M"))
    if center == "mtarchive":
        uri = ("https://mtarchive.geol.iastate.edu/%s/mrms/ncep/%s/%s") % (
            valid.strftime("%Y/%m/%d"),
            product,
            fn,
        )
    else:
        uri = ("https://mrms%s.ncep.noaa.gov/data/2D/%s/MRMS_%s") % (
            center,
            product,
            fn,
        )
    return uri


def fetch(product, valid, tmpdir="/mesonet/tmp"):
    """Get a desired MRMS product

    Applies the following logic:
        - does the file exist in `tmpdir`?
        - can I fetch it from mtarchive?
        - if recent, can I fetch it from NOAA MRMS website

    Args:
      product(str): MRMS product type
      valid(datetime): Datetime object for desired timestamp
      tmpdir(str,optional): location to check/place the downloaded file
    """
    fn = "%s_00.00_%s00.grib2.gz" % (product, valid.strftime("%Y%m%d-%H%M"))
    tmpfn = "%s/%s" % (tmpdir, fn)
    # Option 1, we have this file already in cache!
    if os.path.isfile(tmpfn):
        return tmpfn
    # Option 2, go fetch it from mtarchive
    try:
        req = requests.get(get_url("mtarchive", valid, product), timeout=30)
    except Exception:
        req = None
    if req and req.status_code == 200 and is_gzipped(req.content):
        with open(tmpfn, "wb") as fd:
            fd.write(req.content)
        return tmpfn
    # Option 3, we go look at MRMS website, if timestamp is recent
    utcnow = datetime.utcnow()
    if valid.tzinfo is not None:
        utcnow = utcnow.replace(tzinfo=timezone.utc)
    if (utcnow - valid).total_seconds() > 86400:
        # Can't do option 3!
        return None
    # Loop over all IDP data centers
    for center in ["", "-bldr", "-cprk"]:
        try:
            req = requests.get(get_url(center, valid, product), timeout=30)
        except Exception:
            req = None
        if req and req.status_code == 200 and is_gzipped(req.content):
            with open(tmpfn, "wb") as fd:
                fd.write(req.content)
            return tmpfn
    return None


def make_colorramp():
    """
    Make me a crude color ramp
    """
    c = np.zeros((256, 3), np.int)

    # Ramp blue
    for b in range(0, 37):
        c[b, 2] = 255
    for b in range(37, 77):
        c[b, 2] = (77 - b) * 6
    for b in range(160, 196):
        c[b, 2] = (b - 160) * 6
    for b in range(196, 256):
        c[b, 2] = 254
    # Ramp Green up
    for g in range(0, 37):
        c[g, 1] = g * 6
    for g in range(37, 116):
        c[g, 1] = 254
    for g in range(116, 156):
        c[g, 1] = (156 - g) * 6
    for g in range(196, 256):
        c[g, 1] = (g - 196) * 4
    # and Red
    for r in range(77, 117):
        c[r, 0] = (r - 77) * 6.0
    for r in range(117, 256):
        c[r, 0] = 254

    # Gray for missing
    c[255, :] = [144, 144, 144]
    # Black to remove, eventually
    c[0, :] = [0, 0, 0]
    return tuple(c.ravel())


def reader(fn):
    """ Return metadata and the data """
    fp = gzip.open(fn, "rb")
    metadata = {}
    (
        year,
        month,
        day,
        hour,
        minute,
        second,
        nx,
        ny,
        nz,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        ul_lon_cc,
        ul_lat_cc,
        _,
        scale_lon,
        scale_lat,
        grid_scale,
    ) = struct.unpack("9i4c10i", fp.read(80))

    metadata["ul_lon_cc"] = ul_lon_cc / float(scale_lon)
    metadata["ul_lat_cc"] = ul_lat_cc / float(scale_lat)
    # Calculate
    metadata["ll_lon_cc"] = metadata["ul_lon_cc"]
    metadata["ll_lat_cc"] = metadata["ul_lat_cc"] - (
        (scale_lat / float(grid_scale)) * (ny - 1)
    )
    metadata["ll_lat"] = (
        metadata["ll_lat_cc"] - (scale_lat / float(grid_scale)) / 2.0
    )
    metadata["ul_lat"] = (
        metadata["ul_lat_cc"] - (scale_lat / float(grid_scale)) / 2.0
    )
    metadata["ll_lon"] = (
        metadata["ll_lon_cc"] - (scale_lon / float(grid_scale)) / 2.0
    )
    metadata["ul_lon"] = (
        metadata["ul_lon_cc"] - (scale_lon / float(grid_scale)) / 2.0
    )

    metadata["valid"] = datetime(
        year, month, day, hour, minute, second
    ).replace(tzinfo=timezone.utc)

    struct.unpack("%si" % (nz,), fp.read(nz * 4))  # levels
    struct.unpack("i", fp.read(4))  # z_scale
    struct.unpack("10i", fp.read(40))  # bogus
    struct.unpack("20c", fp.read(20))  # varname
    metadata["unit"] = struct.unpack("6c", fp.read(6))
    var_scale, _, num_radars = struct.unpack("3i", fp.read(12))
    struct.unpack(
        "%sc" % (num_radars * 4,), fp.read(num_radars * 4)
    )  # rad_list
    sz = nx * ny * nz
    data = struct.unpack("%sh" % (sz,), fp.read(sz * 2))
    data = np.reshape(np.array(data), (ny, nx)) / float(var_scale)
    # ma.masked_equal(data, miss_val)

    fp.close()
    return metadata, data


def get_fn(prefix, now, tile):
    """ Get the filename for this timestamp and tile """
    return now.strftime(
        f"/mnt/a4/data/%Y/%m/%d/mrms/tile{tile}/{prefix}/{prefix}"
        ".%Y%m%d.%H%M00.gz"
    )


def write_worldfile(filename):
    """Write a worldfile to the given filename

    Args:
      filename (str): filename to write the world file information to
    """
    with open(filename, "w") as fd:
        fd.write(f"0.01\n0.00\n0.00\n-0.01\n{WEST}\n{NORTH}")



"""Unit Tests"""
# pylint: disable=redefined-outer-name

import pytest
from psycopg2.extras import RealDictCursor
from pyiem.nws.products.spcpts import parser, str2multipolygon, load_conus_data
from pyiem.util import get_dbconn, utc, get_test_file


@pytest.fixture
def dbcursor():
    """Get database."""
    return get_dbconn("postgis").cursor(cursor_factory=RealDictCursor)


def test_issue295_geometryfail():
    """Test that we can make a geometry out of this."""
    # https://.../products/outlook/archive/2020/day1otlk_20200926_1300.html
    s = (
        "45048223 44728372 45018571 44818788 44798832 44758914 "
        "44679048 44759155 45509260 46349268 46749220 47199029 "
        "47608787 47258575"
    )
    load_conus_data(utc(2020, 9, 26))
    res = str2multipolygon(s)
    assert abs(res[0].area - 21.09) < 0.001


def test_200602_unpack():
    """Workaround a full failure, but this still fails :("""
    # https://.../products/outlook/archive/2020/day2otlk_20200602_1730.html
    spc = parser(get_test_file("SPCPTS/PTSDY2_unpack.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT", 2)
    assert abs(outlook.geometry.area - 78.7056) < 0.01


def test_200109_nogeoms():
    """Failed to parse some tricky line work south of New Orleans."""
    # https://.../products/outlook/archive/2020/day2otlk_20200109_1730.html
    spc = parser(get_test_file("SPCPTS/PTSDY2_nogeom3.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "ENH", 2)
    assert abs(outlook.geometry.area - 33.785) < 0.01


def test_190907_invalid():
    """Product hit geos issue."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_190907.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "TSTM", 1)
    assert abs(outlook.geometry.area - 314.76) < 0.01


def test_190905_invalid():
    """Product hit geos issue."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_geos.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "TSTM", 1)
    assert abs(outlook.geometry.area - 263.61) < 0.01


def test_190903_invalid():
    """Product hit invalid geometry error."""
    spc = parser(get_test_file("SPCPTS/PTSDY2_invalid2.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "TSTM", 2)
    assert abs(outlook.geometry.area - 343.74) < 0.01


def test_190801_shapely():
    """Product hit shapely assertion error."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_shapelyerror.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "TSTM", 1)
    assert abs(outlook.geometry.area - 333.678) < 0.01


def test_190625_nogeom2():
    """This hit some error that we need to debug."""
    spc = parser(get_test_file("SPCPTS/PTSDY2_nogeom2.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT", 2)
    assert abs(outlook.geometry.area - 11.59) < 0.01


def test_190527_canada():
    """SPC Updated marine bounds."""
    # https://.../products/outlook/archive/2019/day1otlk_20190528_0100.html
    spc = parser(get_test_file("SPCPTS/PTSDY1_canada.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "MRGL", 1)
    assert abs(outlook.geometry.area - 118.245) < 0.01


def test_190515_issue117_month():
    """Product crossing year causes grief."""
    spc = parser(get_test_file("SPCPTS/PTSDY2_month.txt"))
    collect = spc.get_outlookcollection(2)
    assert collect.expire == utc(2019, 5, 2, 12)


def test_190509_marinebounds():
    """SPC Updated marine bounds."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_marine.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("HAIL", "0.15", 1)
    assert abs(outlook.geometry.area - 17.82) < 0.01


def test_190415_elevated():
    """Can we parse elevated threshold firewx?"""
    spc = parser(get_test_file("SPCPTS/PFWFD1_example.txt"))
    outlook = spc.get_outlook("FIRE WEATHER CATEGORICAL", "ELEV", 1)
    assert abs(outlook.geometry.area - 145.64) < 0.01
    for level in ["IDRT", "SDRT", "ELEV", "CRIT", "EXTM"]:
        outlook = spc.get_outlook("FIRE WEATHER CATEGORICAL", level, 1)
        assert outlook is not None


def test_190415_badtime():
    """This product has a bad time period, we should emit a warning."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_invalidtime.txt"))
    assert any([w.startswith("time_bounds_check") for w in spc.warnings])


def test_180807_idx1_idx2():
    """This Day1 generated an error."""
    spc = parser(get_test_file("SPCPTS/PTSDY1_idx1_idx2.txt"))
    outlook = spc.get_outlook("WIND", "0.05", 1)
    assert abs(outlook.geometry.area - 37.83) < 0.02


def test_170926_largeenh(dbcursor):
    """This Day1 generated a massive ENH"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_bigenh.txt"))
    # spc.draw_outlooks()
    spc.sql(dbcursor)
    outlook = spc.get_outlook("CATEGORICAL", "ENH", 1)
    assert abs(outlook.geometry.area - 17.50) < 0.01


def test_170703_badday3link():
    """Day3 URL is wrong"""
    spc = parser(get_test_file("SPCPTS/PTSDY3.txt"))
    jdict = spc.get_jabbers("", "")
    ans = (
        "The Storm Prediction Center issues Day 3 "
        "Convective Outlook at Nov 19, 8:31z "
        "https://www.spc.noaa.gov/products/outlook/archive/2013/"
        "day3otlk_20131119_0830.html"
    )
    assert jdict[0][0] == ans


def test_170612_nullgeom(dbcursor):
    """See why this has an error with null geom reported"""
    spc = parser(get_test_file("SPCPTS/PTSD48_nullgeom.txt"))
    # spc.draw_outlooks()
    spc.sql(dbcursor)
    outlook = spc.get_outlook("ANY SEVERE", "0.15", 4)
    assert abs(outlook.geometry.area - 56.84) < 0.01


def test_170522_nogeom():
    """See why this has an error with no-geom reported"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_nogeom2.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("TORNADO", "0.02", 1)
    assert abs(outlook.geometry.area - 2.90) < 0.01


def test_170518_bad_dbtime():
    """This went into the database with an incorrect expiration time"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_baddbtime.txt"))
    answer = utc(2017, 5, 1, 12, 0)
    for _, outlook in spc.outlook_collections.items():
        assert outlook.expire == answer


def test_170428_large(dbcursor):
    """PTSDY1 has a large 10 tor"""
    # https://.../products/outlook/archive/2006/day1otlk_20060510_1630.html
    spc = parser(get_test_file("SPCPTS/PTSDY1_largetor10.txt"))
    # spc.draw_outlooks()
    spc.sql(dbcursor)
    outlook = spc.get_outlook("TORNADO", "0.10", 1)
    assert abs(outlook.geometry.area - 31.11) < 0.01
    outlook = spc.get_outlook("CATEGORICAL", "TSTM", 1)
    assert abs(outlook.geometry.area - 428.00) < 0.01


def test_170417_empty(dbcursor):
    """An empty PTSD48 was causing an exception in get_jabbers"""
    spc = parser(get_test_file("SPCPTS/PTSD48_empty.txt"))
    # spc.draw_outlooks()
    spc.sql(dbcursor)
    jabber = spc.get_jabbers("")
    ans = (
        "The Storm Prediction Center issues Days 4-8 "
        "Convective Outlook at Dec 25, 9:41z "
        "https://www.spc.noaa.gov/products/exper/day4-8/archive/"
        "2008/day4-8_20081225.html"
    )
    assert jabber[0][0] == ans


def test_051128_invalid(dbcursor):
    """Make sure that the SIG wind threshold does not eat the US"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_biggeom2.txt"))
    # spc.draw_outlooks()
    spc.sql(dbcursor)
    # Both of these are invalidly provided in the PTS file and should be
    # dumped as they are larger than the general thunder
    outlook = spc.get_outlook("WIND", "SIGN", 1)
    assert outlook.geometry.is_empty
    outlook = spc.get_outlook("WIND", "0.05", 1)
    assert outlook.geometry.is_empty
    print("\n".join(spc.warnings))
    assert len(spc.warnings) == 2


def test_080731_invalid():
    """Make sure that the SIG wind threshold does not eat the US"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_biggeom.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("WIND", "SIGN", 1)
    assert abs(outlook.geometry.area - 15.82) < 0.01
    assert len(spc.warnings) == 1


def test_170411_jabber_error():
    """This empty Fire Weather Day 3-8 raised a jabber error"""
    spc = parser(get_test_file("SPCPTS/PFWF38_empty.txt"))
    j = spc.get_jabbers("")
    ans = (
        "The Storm Prediction Center issues Day 3-8 Fire "
        "Weather Outlook at Apr 11, 19:54z "
        "https://www.spc.noaa.gov/products/exper/fire_wx/2017/170413.html"
    )
    assert j[0][0] == ans


def test_170406_day48_pre2015(dbcursor):
    """Can we parse a pre2015 days 4-8"""
    spc = parser(get_test_file("SPCPTS/PTSD48_pre2015.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("ANY SEVERE", "0.15", 4)
    assert abs(outlook.geometry.area - 73.116) < 0.01
    outlook = spc.get_outlook("ANY SEVERE", "0.15", 5)
    assert abs(outlook.geometry.area - 72.533) < 0.01
    spc.sql(dbcursor)


def test_170406_day48(dbcursor):
    """Can we parse a present day days 4-8"""
    spc = parser(get_test_file("SPCPTS/PTSD48.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("ANY SEVERE", "0.15", 4)
    assert abs(outlook.geometry.area - 40.05) < 0.01
    spc.sql(dbcursor)
    collect = spc.get_outlookcollection(4)
    assert collect.issue == utc(2017, 4, 9, 12)
    assert collect.expire == utc(2017, 4, 10, 12)


def test_170404_nogeom():
    """nogeom error from a 2002 product"""
    # 26 Sep 2017, we can workaround this now
    spc = parser(get_test_file("SPCPTS/PTSDY1_2002_nogeom.txt"))
    outlook = spc.get_outlook("TORNADO", "0.05")
    assert abs(outlook.geometry.area - 8.76) < 0.01


def test_170404_2002():
    """Can we parse something from 2002?"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_2002.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 38.614) < 0.01


def test_170329_notimp():
    """Exception was raised parsing this guy"""
    spc = parser(get_test_file("SPCPTS/PTSDY2_notimp.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "MRGL")
    assert abs(outlook.geometry.area - 110.24) < 0.01


def test_170215_gh23():
    """A marginal for the entire country :/"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_gh23.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "MRGL")
    assert abs(outlook.geometry.area - 19.63) < 0.01


def test_150622_ptsdy1_topo():
    """PTSDY1_topo.txt """
    spc = parser(get_test_file("SPCPTS/PTSDY1_topo.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 91.91) < 0.01


def test_150622_ptsdy2():
    """PTSDY2_invalid.txt parsed ok."""
    spc = parser(get_test_file("SPCPTS/PTSDY2_invalid.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 78.14) < 0.01


def test_150622_ptsdy1():
    """PTSDY1_nogeom.txt """
    spc = parser(get_test_file("SPCPTS/PTSDY1_nogeom.txt"))
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 95.900) < 0.01


def test_150612_ptsdy1_3():
    """We got an error with this, so we shall test"""
    spc = parser(get_test_file("SPCPTS/PTSDY1_3.txt"))
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 53.94) < 0.01


def test_141022_newcats():
    """ Make sure we can parse the new categories """
    spc = parser(
        get_test_file("SPCPTS/PTSDY1_new.txt"),
        utcnow=utc(2014, 10, 13, 16, 21),
    )
    outlook = spc.get_outlook("CATEGORICAL", "ENH")
    assert abs(outlook.geometry.area - 13.02) < 0.01
    outlook = spc.get_outlook("CATEGORICAL", "MRGL")
    assert abs(outlook.geometry.area - 47.01) < 0.01


def test_140709_nogeoms():
    """Can we parse holes."""
    spc = parser(get_test_file("SPCPTS/PTSDY3_nogeoms.txt"))
    outlook = spc.get_outlook("ANY SEVERE", "0.05")
    assert abs(outlook.geometry.area - 99.68) < 0.01


def test_140710_nogeom():
    """Can we parse holes."""
    spc = parser(get_test_file("SPCPTS/PTSDY2_nogeom.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("CATEGORICAL", "SLGT")
    assert abs(outlook.geometry.area - 43.02) < 0.01


def test_23jul_failure():
    """ CCW line near Boston """
    # need to load data for this to work as a one
    load_conus_data(utc(2017, 7, 23))
    data = """40067377 40567433 41317429 42097381 42357259 42566991"""
    res = str2multipolygon(data)
    assert abs(res[0].area - 7.96724) < 0.0001


def test_140707_general():
    """ Had a problem with General Thunder, lets test this """
    # https://.../products/outlook/archive/2014/day1otlk_20140707_1630.html
    spc = parser(get_test_file("SPCPTS/PTSDY1_complex.txt"))
    # spc.draw_outlooks()
    # Linework here is invalid, so we can't account for it.
    outlook = spc.get_outlook("CATEGORICAL", "TSTM")
    assert abs(outlook.geometry.area - 755.424) < 0.01


def test_complex():
    """ Test our processing """
    spc = parser(get_test_file("SPCPTS/PTSDY3.txt"))
    outlook = spc.get_outlook("ANY SEVERE", "0.05")
    assert abs(outlook.geometry.area - 10.12) < 0.01


def test_bug_140601_pfwf38():
    """ Encounted issue with Fire Outlook Day 3-8 """
    spc = parser(get_test_file("SPCPTS/PFWF38.txt"))
    # spc.draw_outlooks()
    collect = spc.get_outlookcollection(3)
    assert len(collect.outlooks) == 1


def test_bug_140507_day1():
    """ Bug found in production with GEOS Topology Exception """
    spc = parser(get_test_file("SPCPTS/PTSDY1_topoexp.txt"))
    # spc.draw_outlooks()
    collect = spc.get_outlookcollection(1)
    assert len(collect.outlooks) == 14


def test_bug_140506_day2():
    """Bug found in production"""
    spc = parser(get_test_file("SPCPTS/PTSDY2.txt"))
    # spc.draw_outlooks()
    collect = spc.get_outlookcollection(2)
    assert len(collect.outlooks) == 6
    j = spc.get_jabbers("localhost", "localhost")
    ans = (
        "The Storm Prediction Center issues Day 2 "
        "Convective Outlook at May 6, 17:31z "
        "https://www.spc.noaa.gov/products/outlook/archive/2014/"
        "day2otlk_20140506_1730.html"
    )
    assert j[0][0] == ans


def test_bug_140518_day2():
    """ 18 May 2014 tripped error with no exterior polygon found """
    spc = parser(get_test_file("SPCPTS/PTSDY2_interior.txt"))
    # spc.draw_outlooks()
    collect = spc.get_outlookcollection(2)
    assert len(collect.outlooks) == 1


def test_bug_140519_day1():
    """ 19 May 2014 tripped error with no exterior polygon found """
    spc = parser(get_test_file("SPCPTS/PTSDY1_interior.txt"))
    # spc.draw_outlooks()
    collect = spc.get_outlookcollection(1)
    assert len(collect.outlooks) == 7


def test_bug():
    """ Test bug list index outof range """
    spc = parser(get_test_file("SPCPTS/PTSDY1_2.txt"))
    collect = spc.get_outlookcollection(1)
    assert len(collect.outlooks) == 1


def test_complex_2():
    """ Test our processing """
    spc = parser(get_test_file("SPCPTS/PTSDY1.txt"))
    # spc.draw_outlooks()
    outlook = spc.get_outlook("HAIL", "0.05")
    assert abs(outlook.geometry.area - 47.65) < 0.01


def test_str1(dbcursor):
    """ check spcpts parsing """
    spc = parser(get_test_file("SPCPTS/SPCPTS.txt"))
    # spc.draw_outlooks()
    assert spc.valid == utc(2013, 7, 19, 19, 52)
    assert spc.issue == utc(2013, 7, 19, 20, 0)
    assert spc.expire == utc(2013, 7, 20, 12, 0)

    spc.sql(dbcursor)
    spc.compute_wfos(dbcursor)
    # It is difficult to get a deterministic result here as in Travis, we
    # don't have UGCS, so the WFO lookup yields no results
    j = spc.get_jabbers("")
    assert len(j) >= 1



"""Test our CF6 parsing."""
# pylint: disable=redefined-outer-name
import datetime

import pytest
from pyiem.nws.products.cf6 import parser
from pyiem.util import get_test_file, get_dbconn
from pyiem.reference import TRACE_VALUE


@pytest.fixture
def dbcursor():
    """Return a database cursor."""
    pgconn = get_dbconn("iem")
    yield pgconn.cursor()
    pgconn.close()


def test_201226_bad_date(dbcursor):
    """Test that no error is emitted for a CF6 with a 'Bad Date'."""
    prod = parser(get_test_file("CF6/CF6WYS_error.txt"))
    prod.sql(dbcursor)
    assert prod


def test_200421_nan(dbcursor):
    """Test database insert that was failing with NaN values."""
    prod = parser(get_test_file("CF6/CF6MKK.txt"))
    prod.sql(dbcursor)

    dbcursor.execute(
        "SELECT wxcodes from cf6_data_2020 where station = 'PMKK' and "
        "valid = '2020-04-18'"
    )
    assert dbcursor.fetchone()[0] == "X"


def test_200302_regex_error():
    """Test failure found with some regex failure."""
    prod = parser(get_test_file("CF6/CF6GRR.txt"))
    assert prod.df.iloc[0]["max"] == 51


def test_200224_time():
    """Test failure found with timestamp parsing."""
    prod = parser(get_test_file("CF6/CF6WYS.txt"))
    assert prod.df.iloc[0]["max"] == 32


def test_basic(dbcursor):
    """Test CF6 Parsing."""
    prod = parser(get_test_file("CF6/CF6DSM.txt"))
    assert prod.df.iloc[0]["max"] == 42
    prod.sql(dbcursor)


def test_missing_header():
    """Test exception when there is a missing header."""
    with pytest.raises(ValueError):
        parser(get_test_file("CF6/CF6DSM_bad.txt"))


def test_nodata(dbcursor):
    """Test when there is no data in the product."""
    prod = parser(get_test_file("CF6/CF6DSM_empty.txt"))
    prod.sql(dbcursor)
    assert prod.df.empty


def test_trace(dbcursor):
    """Ensure that our decoder is properly dealing with trace values."""
    prod = parser(get_test_file("CF6/CF6SEA.txt"))
    assert prod.df.index.values[0] == datetime.date(2020, 2, 1)
    assert prod.df.iloc[15]["wtr"] == TRACE_VALUE
    prod.sql(dbcursor)

    dbcursor.execute(
        "SELECT possible_sunshine from cf6_data_2020 where station = 'KSEA' "
        "and valid = '2020-02-01'"
    )
    assert dbcursor.fetchone()[0] is None
